from __future__ import annotations

from enum import Enum, auto
from io import StringIO
from typing import TYPE_CHECKING, Any, Iterable, Literal, Mapping
from dataclasses import dataclass
import numpy as np
from numpy.typing import NDArray

from qtpy import QtWidgets as QtW
from qtpy import QtGui, QtCore
from qtpy.QtCore import Qt

from himena.consts import StandardType
from himena.qt._utils import get_main_window
from himena.style import Theme
from himena.types import WidgetDataModel
from himena.standards.model_meta import TableMeta
from himena.qt._qsvg import QColoredSVGIcon
from himena.plugins import validate_protocol, config_field
from himena_builtins._consts import ICON_PATH
from himena_builtins.qt.widgets._table_components import (
    QTableBase,
    QSelectionRangeEdit,
    FLAGS,
    Editability,
)
from himena.utils.collections import UndoRedoStack
from himena.utils.misc import table_to_text

if TYPE_CHECKING:
    from himena.widgets import MainWindow


class HeaderFormat(Enum):
    """Enum of how to index table header."""

    NumberZeroIndexed = auto()
    NumberOneIndexed = auto()
    Alphabetic = auto()


@dataclass
class TableAction:
    """Base class for table undo/redo actions."""

    def invert(self) -> TableAction:
        return self

    def apply(self, table: QSpreadsheet):
        raise NotImplementedError("Apply method must be implemented.")


@dataclass
class EditAction(TableAction):
    old: str | np.ndarray
    new: str | np.ndarray
    index: tuple[int | slice, int | slice]

    def invert(self) -> TableAction:
        return EditAction(self.new, self.old, self.index)

    def apply(self, table: QSpreadsheet):
        return table.array_update(self.index, self.new, record_undo=False)


@dataclass
class ReshapeAction(TableAction):
    old: tuple[int, int]
    new: tuple[int, int]

    def invert(self) -> TableAction:
        return ReshapeAction(self.new, self.old)

    def apply(self, table: QSpreadsheet):
        r_old, c_old = self.old
        r_new, c_new = self.new
        if r_old == r_new and c_old == c_new:
            pass
        elif r_old < r_new and c_old < c_new:
            table.array_expand(r_new, c_new)
        elif r_old > r_new and c_old > c_new:
            table.array_shrink(r_new, r_new)
        else:
            raise ValueError(
                f"This reshape ({self.old} -> {self.new}) is not supported."
            )


@dataclass
class InsertAction(TableAction):
    index: int
    axis: Literal[0, 1]
    array: np.ndarray | None = None

    def invert(self) -> TableAction:
        return RemoveAction(self.index, self.axis, self.array)

    def apply(self, table: QSpreadsheet):
        table.array_insert(self.index, self.axis, self.array, record_undo=False)


@dataclass
class RemoveAction(TableAction):
    index: int
    axis: Literal[0, 1]
    array: np.ndarray

    def invert(self) -> TableAction:
        return InsertAction(self.index, self.axis, self.array)

    def apply(self, table: QSpreadsheet):
        table.array_delete([self.index], self.axis, record_undo=False)


@dataclass
class ActionGroup(TableAction):
    actions: list[TableAction]  # operation from actions[0] to actions[-1]

    def invert(self) -> TableAction:
        return ActionGroup([action.invert() for action in self.actions[::-1]])

    def apply(self, table: QSpreadsheet):
        for action in self.actions:
            action.apply(table)


class QStringArrayModel(QtCore.QAbstractTableModel):
    """Table model for a string array."""

    MIN_ROW_COUNT = 100
    MIN_COLUMN_COUNT = 30

    def __init__(self, arr: np.ndarray, parent: QSpreadsheet):
        super().__init__(parent)
        self._arr = arr  # 2D
        if arr.ndim != 2:
            raise ValueError("Only 2D array is supported.")
        if not isinstance(arr.dtype, np.dtypes.StringDType):
            raise ValueError("Only string array is supported.")
        self._nrows, self._ncols = arr.shape
        self._is_original_array = True
        self._header_format = HeaderFormat.NumberZeroIndexed

    @classmethod
    def empty(cls, parent: QSpreadsheet) -> QStringArrayModel:
        return cls(np.empty((0, 0), dtype=np.dtypes.StringDType()), parent)

    if TYPE_CHECKING:

        def parent(self) -> QSpreadsheet: ...  # fmt: skip

    def rowCount(self, parent=None):
        return max(self._nrows + 1, self.MIN_ROW_COUNT)

    def columnCount(self, parent=None):
        return max(self._ncols + 1, self.MIN_COLUMN_COUNT)

    def set_array(self, arr: np.ndarray, is_original: bool = True) -> None:
        if arr.ndim != 2:
            raise ValueError("Only 2D array is supported.")
        nr, nc = arr.shape
        nr0, nc0 = self.rowCount(), self.columnCount()
        self._arr = arr
        self._nrows, self._ncols = arr.shape

        # adjust the model size to fit the new array.
        _index = QtCore.QModelIndex()
        if nr + 1 > nr0:
            self.beginInsertRows(_index, nr0, nr)
            self.insertRows(nr0, nr + 1 - nr0, _index)
            self.endInsertRows()
        elif nr + 1 < nr0:
            nr_next = max(self.MIN_ROW_COUNT, nr + 1)
            self.beginRemoveRows(_index, nr_next, nr0 - 1)
            self.removeRows(nr_next, nr0 - nr_next - 2, _index)
            self.endRemoveRows()
        if nc + 1 > nc0:
            self.beginInsertColumns(_index, nc0, nc)
            self.insertColumns(nc0, nc + 1 - nc0, _index)
            self.endInsertColumns()
        elif nc + 1 < nc0:
            nc_next = max(self.MIN_COLUMN_COUNT, nc + 1)
            self.beginRemoveColumns(_index, nc_next, nc0 - 1)
            self.removeColumns(nc_next, nc0 - nc_next - 2, _index)
            self.endRemoveColumns()
        if not (is_original and self._is_original_array):
            # no need for further copy-on-write
            self._is_original_array = False

    def flags(self, index: QtCore.QModelIndex) -> Qt.ItemFlag:
        return FLAGS

    def data(
        self,
        index: QtCore.QModelIndex,
        role: Qt.ItemDataRole = Qt.ItemDataRole.DisplayRole,
    ):
        if not index.isValid():
            return QtCore.QVariant()
        r, c = index.row(), index.column()
        if r >= self._arr.shape[0] or c >= self._arr.shape[1]:
            return QtCore.QVariant()
        if role == Qt.ItemDataRole.TextAlignmentRole:
            return Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter
        if role == Qt.ItemDataRole.ToolTipRole:
            return f"A[{r}, {c}] = {self._arr[r, c]}"
        if role in [Qt.ItemDataRole.EditRole, Qt.ItemDataRole.DisplayRole]:
            return str(self._arr[r, c])
        return QtCore.QVariant()

    def setData(self, index: QtCore.QModelIndex, value: Any, role: int = ...) -> bool:
        if role == Qt.ItemDataRole.EditRole:
            qtable = self.parent()
            qtable.array_update((index.row(), index.column()), value, record_undo=True)
            return True
        return False

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ):
        if role != Qt.ItemDataRole.DisplayRole:
            return QtCore.QVariant()
        if self._header_format is HeaderFormat.NumberZeroIndexed:
            return str(section)
        elif self._header_format is HeaderFormat.NumberOneIndexed:
            return str(section + 1)
        elif self._header_format is HeaderFormat.Alphabetic:
            if orientation == Qt.Orientation.Horizontal:
                return str(char_arange(section, section + 1)[0])
            else:
                return str(section + 1)


SEP_NAMES = {"\t": "Tab", " ": "Space", ",": "Comma", ";": "Semicolon"}


class QSpreadsheet(QTableBase):
    """Table widget for editing a 2D string array.

    ## Basic Usage

    Moving, selecting, editing and copying/pasting cells are supported like many other
    spreadsheet applications.

    ## Keyboard Shortcuts

    - `↑`, `↓`, `←`, `→`: Move the current cell.
    - `PgUp`, `PgDn`, `Home`, `End`: Move the current cell by 10.
    - `Ctrl+↑`, `Ctrl+↓`, `Ctrl+←`, `Ctrl+→`: Move the current cell to the edge.
    - `Shift` with the above keys: Extend the selection.
    - `Ctrl+X`: Cut the selected cells.
    - `Ctrl+C`: Copy the selected cells to the clipboard.
    - `Ctrl+V`: Paste the clipboard content to the selected cells.
    - `Ctrl+Z`: Undo the last action.
    - `Ctrl+Y`: Redo the last undone action.
    - `Ctrl+A`: Select all cells.
    - `F2`: Start editing the current cell.
    - `Ctrl+F`: Find a string in the table.

    ## Mouse Interaction

    - Click on a cell to select it.
    - Drag to select multiple cells.
    - Double click on a cell to edit it.
    - Right click for a context menu.
    - Right drag to scroll the table.
    """

    __himena_widget_id__ = "builtins:QSpreadsheet"
    __himena_display_name__ = "Built-in Spreadsheet Editor"
    HeaderFormat = HeaderFormat

    def __init__(self, ui: MainWindow):
        QTableBase.__init__(self, ui)
        self.setEditTriggers(Editability.TRUE)
        self._control = None
        self._model_type = StandardType.TABLE
        self._undo_stack = UndoRedoStack[TableAction](size=25)
        self._sep_on_copy = "\t"
        self._extension_default = ".csv"
        self.setModel(QStringArrayModel.empty(self))

    def setHeaderFormat(self, value: HeaderFormat) -> None:
        if model := self.model():
            model._header_format = value

    def data_shape(self) -> tuple[int, int]:
        return self.model()._arr.shape

    @validate_protocol
    def update_model(self, model: WidgetDataModel) -> None:
        value = model.value
        if value is None:
            table = np.empty((0, 0), dtype=np.dtypes.StringDType())
        else:
            if isinstance(value, Mapping):
                table = _dict_to_array(value)
            else:
                table = _array_like_to_array(value)
            if table.ndim < 2:
                table = table.reshape(-1, 1)
        if self.model() is None:
            self.setModel(QStringArrayModel(table, self))
        else:
            self.model().set_array(table)
        sep: str | None = None
        if isinstance(meta := model.metadata, TableMeta):
            if meta.separator is not None:
                sep = meta.separator
            if (pos := meta.current_position) is not None:
                index = self.model().index(*pos)
                self.setCurrentIndex(index)
                self._selection_model.current_index = pos
            if meta.selections:  # if has selections, they need updating
                self._selection_model.clear()
            for (r0, r1), (c0, c1) in meta.selections:
                self._selection_model.append((slice(r0, r1), slice(c0, c1)))

        self._undo_stack.clear()
        self._modified_override = None
        self.update()

        # update control widget
        if self._control is None:
            self._control = QTableControl(self)
        self._control.update_for_table(self)
        if sep is not None:
            self._control._separator_label.setText(f"Sep: {SEP_NAMES.get(sep, sep)}")
            self._control._separator = sep
            self._control._separator_label.show()
        else:
            self._control._separator = None
            self._control._separator_label.hide()
        self._model_type = model.type
        if ext := model.extension_default:
            self._extension_default = ext
        return None

    @validate_protocol
    def to_model(self) -> WidgetDataModel[np.ndarray]:
        meta = self._prep_table_meta()
        if sep := self._control._separator:
            meta.separator = sep
        # NOTE: if this model is passed to another widget and modified in this widget,
        # the other array will be modified as well. To avoid this, we need to reset the
        # copy-on-write state of this array.
        self.model()._is_original_array = True
        return WidgetDataModel(
            value=self.model()._arr,
            type=self.model_type(),
            extension_default=self._extension_default,
            metadata=meta,
        )

    @validate_protocol
    def model_type(self):
        return self._model_type

    @validate_protocol
    def update_configs(self, cfg: SpreadsheetConfigs):
        self.horizontalHeader().setDefaultSectionSize(cfg.default_cell_width)
        self.verticalHeader().setDefaultSectionSize(cfg.default_cell_height)
        self._sep_on_copy = cfg.separator_on_copy.encode().decode("unicode_escape")

    @validate_protocol
    def is_modified(self) -> bool:
        if self._modified_override is not None:
            return self._modified_override
        return self._undo_stack.undoable()

    @validate_protocol
    def set_modified(self, value: bool) -> None:
        self._modified_override = value

    @validate_protocol
    def is_editable(self) -> bool:
        return self.editTriggers() == Editability.TRUE

    @validate_protocol
    def set_editable(self, value: bool) -> None:
        if value:
            trig = Editability.TRUE
        else:
            trig = Editability.FALSE
        self.setEditTriggers(trig)

    @validate_protocol
    def control_widget(self) -> QTableControl:
        return self._control

    @validate_protocol
    def theme_changed_callback(self, theme: Theme) -> None:
        if self._control:
            self._control.update_theme(theme)

    def array_update(
        self,
        index: tuple[int | slice, int | slice],
        value: str,
        *,
        record_undo: bool = True,
    ) -> None:
        """Update the data at the given index."""
        r, c = index
        arr = self.model()._arr
        _ud_old_shape = arr.shape
        r_max = r.stop - 1 if isinstance(r, slice) else r
        c_max = c.stop - 1 if isinstance(c, slice) else c
        if r_max >= arr.shape[0] or c_max >= arr.shape[1]:  # need expansion
            _ud_old_data = ""
            _ud_old_shape = arr.shape
            self.array_expand(r_max + 1, c_max + 1)
            _ud_new_shape = arr.shape
            _action_reshape = ReshapeAction(_ud_old_shape, _ud_new_shape)
            arr = self.model()._arr
        else:
            if isinstance(r, slice) or isinstance(c, slice):
                _ud_old_data = arr[r, c].copy()
            else:
                _ud_old_data = arr[r, c]
            _action_reshape = None
        if self.model()._is_original_array:
            self.model()._arr = arr = arr.copy()
        arr[r, c] = value
        _ud_new_data = arr[r, c]
        _action = EditAction(_ud_old_data, _ud_new_data, (r, c))
        if _action_reshape is not None:
            _action = ActionGroup([_action_reshape, _action])
        if record_undo:
            self._undo_stack.push(_action)

    def array_expand(self, nr: int, nc: int):
        """Expand the array to the given shape (nr, nc)."""
        # ReshapeAction must be recorded outside this function.
        old_arr = self.model()._arr
        nr0, nc0 = old_arr.shape
        new_arr = np.pad(
            old_arr,
            [(0, max(nr - nr0, 0)), (0, max(nc - nc0, 0))],
            mode="constant",
            constant_values="",
        )
        self.model().set_array(new_arr, is_original=False)
        self._control.update_for_table(self)
        self.update()

    def array_shrink(self, nr: int, nc: int):
        """Shrink the array to the given shape."""
        # slicing returns the view of the array, so it should be marked as original.
        self.model().set_array(self.model()._arr[:nr, :nc], is_original=True)
        self._control.update_for_table(self)
        self.update()

    def array_insert(
        self,
        index: int,
        axis: Literal[0, 1],
        values: np.ndarray | None = None,
        *,
        record_undo: bool = True,
    ) -> None:
        """Insert an empty array at the given index."""
        self.model().set_array(
            np.insert(
                self.model()._arr,
                index,
                "" if values is None else values,
                axis=axis,
            ),
            is_original=False,
        )
        if record_undo:
            self._undo_stack.push(InsertAction(index, axis, values))
        self.update()

    def array_delete(
        self,
        indices: Iterable[int],
        axis: Literal[0, 1],
        *,
        record_undo: bool = True,
    ):
        """Remove the array at the given index."""
        # Make action group that remove the row/column one by one. Here, indices may be
        # out of range, as this widget is a spreadsheet.
        size_of_axis = self.model()._arr.shape[axis]
        indices = list(indices)
        _action = ActionGroup(
            [
                RemoveAction(idx, axis, self.model()._arr[_sl(idx, axis)].copy())
                for idx in sorted(indices, reverse=True)
                if idx < size_of_axis
            ]
        )
        # Update the underlying array data and redraw the table.
        self.model().set_array(
            np.delete(self.model()._arr, list(indices), axis=axis), is_original=False
        )
        self.update()
        # Record the action if necessary.
        if record_undo:
            self._undo_stack.push(_action)

    def undo(self):
        """Undo the last action."""
        if action := self._undo_stack.undo():
            action.invert().apply(self)
            self.update()

    def redo(self):
        """Redo the last undone action."""
        if action := self._undo_stack.redo():
            action.apply(self)
            self.update()

    def _make_context_menu(self):
        menu = QtW.QMenu(self)
        menu.addAction("Cut", self._cut_and_copy_to_clipboard)
        menu.addAction("Copy", self._copy_to_clipboard)
        copy_as_menu = menu.addMenu("Copy As ...")
        copy_as_menu.addAction("CSV", self._copy_as_csv)
        copy_as_menu.addAction("Markdown", self._copy_as_markdown)
        copy_as_menu.addAction("HTML", self._copy_as_html)
        copy_as_menu.addAction("rST", self._copy_as_rst)
        menu.addAction("Paste", self._paste_from_clipboard)
        menu.addSeparator()
        menu.addAction("Insert Row Above", self._insert_row_above)
        menu.addAction("Insert Row Below", self._insert_row_below)
        menu.addAction("Insert Column Left", self._insert_column_left)
        menu.addAction("Insert Column Right", self._insert_column_right)
        menu.addSeparator()
        menu.addAction("Remove Selected Rows", self._remove_selected_rows)
        menu.addAction("Remove Selected Columns", self._remove_selected_columns)
        menu.addSeparator()
        menu.addAction("Measure At Selection", self._measure)
        return menu

    def _cut_and_copy_to_clipboard(self):
        self._copy_to_clipboard()
        self._delete_selection()

    def _copy_to_clipboard(self, format="TSV"):
        sels = self._selection_model.ranges
        if len(sels) != 1:
            return
        sel = sels[0]
        values = self.model()._arr[sel]
        if values.size > 0:
            string = table_to_text(values, format=format)[0]
            QtW.QApplication.clipboard().setText(string)

    def _copy_as_csv(self):
        return self._copy_to_clipboard("CSV")

    def _copy_as_markdown(self):
        return self._copy_to_clipboard("Markdown")

    def _copy_as_html(self):
        return self._copy_to_clipboard("HTML")

    def _copy_as_rst(self):
        return self._copy_to_clipboard("rST")

    def _paste_from_clipboard(self):
        text = QtW.QApplication.clipboard().text()
        if not text:
            return

        buf = StringIO(text)
        arr_paste = np.loadtxt(
            buf, dtype=np.dtypes.StringDType(), delimiter="\t", ndmin=2
        )
        # undo info
        _ud_old_shape = self.data_shape()

        sl0, sl1, _ud_old_data = self._paste_array(arr_paste)

        # undo info
        _ud_new_shape = self.data_shape()
        _ud_new_data = arr_paste.copy()
        _action_edit = EditAction(_ud_old_data, _ud_new_data, (sl0, sl1))
        if _ud_old_shape == _ud_new_shape:
            _action = _action_edit
        else:
            _action_reshape = ReshapeAction(_ud_old_shape, _ud_new_shape)
            _action = ActionGroup([_action_reshape, _action_edit])
        self._undo_stack.push(_action)

    def _paste_array(self, arr_paste: np.ndarray) -> tuple[slice, slice, np.ndarray]:
        """Update the array and return the pasteed range and old data."""
        arr = self.model()._arr
        # paste in the text
        rng = self._selection_model.get_single_range()
        row0, col0 = rng[0].start, rng[1].start
        lr = max(arr_paste.shape[0], rng[0].stop - rng[0].start)
        lc = max(arr_paste.shape[1], rng[1].stop - rng[1].start)

        # expand the table if necessary
        expanded = False
        if (row0 + lr) > arr.shape[0]:
            arr = np.pad(
                arr,
                [(0, row0 + lr - arr.shape[0]), (0, 0)],
                mode="constant",
                constant_values="",
            )
            expanded = True
        if (col0 + lc) > arr.shape[1]:
            arr = np.pad(
                arr,
                [(0, 0), (0, col0 + lc - arr.shape[1])],
                mode="constant",
                constant_values="",
            )
            expanded = True
        if not expanded:
            arr = arr.copy()
        # paste the data
        target_slice = (slice(row0, row0 + lr), slice(col0, col0 + lc))
        old_data = arr[target_slice].copy()
        arr[target_slice] = arr_paste
        self.model().set_array(arr, is_original=False)

        # select what was just pasted
        self._selection_model.set_ranges([target_slice])
        self.update()
        return target_slice + (old_data,)

    def _delete_selection(self):
        _actions = []
        _maybe_empty_edges = False
        arr = self.model()._arr
        # replace all the selected cells with empty strings.
        for sel in self._selection_model.ranges:
            old_array = arr[sel].copy()
            new_array = np.zeros_like(old_array)
            _actions.append(EditAction(old_array, new_array, sel))
            arr[sel] = ""
            if sel[0].stop == arr.shape[0] or sel[1].stop == arr.shape[1]:
                _maybe_empty_edges = True
        # if this deletion makes the array edges empty, array should be shrunk.
        if _maybe_empty_edges:
            arr_nchars = np.char.str_len(arr)
            size_0 = _size_to_shrink(np.max(arr_nchars, axis=1))
            size_1 = _size_to_shrink(np.max(arr_nchars, axis=0))
            if size_0 < arr_nchars.shape[0] or size_1 < arr_nchars.shape[1]:
                self.array_shrink(size_0, size_1)
                _actions.append(ReshapeAction(arr_nchars.shape, (size_0, size_1)))
        self.update()
        self._undo_stack.push(ActionGroup(_actions))

    def _insert_row_below(self):
        """Insert a row below the current selection."""
        self.array_insert(self._selection_model.current_index.row + 1, 0)

    def _insert_row_above(self):
        """Insert a row above the current selection."""
        self.array_insert(self._selection_model.current_index.row, 0)

    def _insert_column_right(self):
        """Insert a column to the right of the current selection."""
        self.array_insert(self._selection_model.current_index.column + 1, 1)

    def _insert_column_left(self):
        """Insert a column to the left of the current selection."""
        self.array_insert(self._selection_model.current_index.column, 1)

    def _remove_selected_rows(self):
        """Remove the selected rows."""
        selected_rows = set[int]()
        for sel in self._selection_model.ranges:
            selected_rows.update(range(sel[0].start, sel[0].stop))
        self.array_delete(selected_rows, axis=0)

    def _remove_selected_columns(self):
        """Remove the selected columns."""
        selected_cols = set[int]()
        for sel in self._selection_model.ranges:
            selected_cols.update(range(sel[1].start, sel[1].stop))
        self.array_delete(selected_cols, axis=1)

    def _measure(self):
        ui = get_main_window(self)
        ui.exec_action("builtins:table:measure-selection")

    def keyPressEvent(self, e: QtGui.QKeyEvent):
        _Ctrl = QtCore.Qt.KeyboardModifier.ControlModifier
        if e.modifiers() & _Ctrl and e.key() == QtCore.Qt.Key.Key_C:
            return self._copy_to_clipboard()
        elif e.modifiers() & _Ctrl and e.key() == QtCore.Qt.Key.Key_V:
            return self._paste_from_clipboard()
        elif e.modifiers() & _Ctrl and e.key() == QtCore.Qt.Key.Key_X:
            return self._cut_and_copy_to_clipboard()
        elif e.key() in (QtCore.Qt.Key.Key_Delete, QtCore.Qt.Key.Key_Backspace):
            return self._delete_selection()
        elif e.modifiers() & _Ctrl and e.key() == QtCore.Qt.Key.Key_F:
            self._find_string()
            return
        elif e.modifiers() & _Ctrl and e.key() == QtCore.Qt.Key.Key_Z:
            self.undo()
            return
        elif e.modifiers() & _Ctrl and e.key() == QtCore.Qt.Key.Key_Y:
            self.redo()
            return
        elif (
            (not (e.modifiers() & _Ctrl))
            and (
                e.modifiers() == QtCore.Qt.KeyboardModifier.NoModifier
                or e.modifiers() & QtCore.Qt.KeyboardModifier.ShiftModifier
            )
            and Qt.Key.Key_Space <= e.key() <= Qt.Key.Key_ydiaeresis
        ):
            index = self._selection_model.current_index
            qindex = self.model().index(index.row, index.column)
            if qindex.isValid():
                self.edit(qindex)
                if editor := self.itemDelegate()._current_editor_ref():
                    editor.setText(e.text())
                return
        return super().keyPressEvent(e)

    if TYPE_CHECKING:

        def model(self) -> QStringArrayModel: ...


_R_CENTER = QtCore.Qt.AlignmentFlag.AlignRight | QtCore.Qt.AlignmentFlag.AlignVCenter


class QTableControl(QtW.QWidget):
    def __init__(self, table: QSpreadsheet):
        super().__init__()
        layout = QtW.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setAlignment(_R_CENTER)
        self._info_label = QtW.QLabel("")
        self._info_label.setAlignment(_R_CENTER)

        # toolbuttons
        groupbox_ins = QToolButtonGroup(self)
        groupbox_rem = QToolButtonGroup(self)
        _btn_ins = groupbox_ins.add_widgets(
            _tool_btn(table._insert_row_above, "row_insert_top"),
            _tool_btn(table._insert_row_below, "row_insert_bottom"),
            _tool_btn(table._insert_column_left, "col_insert_left"),
            _tool_btn(table._insert_column_right, "col_insert_right"),
        )
        _btn_rem = groupbox_rem.add_widgets(
            _tool_btn(table._remove_selected_rows, "row_remove"),
            _tool_btn(table._remove_selected_columns, "col_remove"),
        )
        self._tool_buttons: list[QtW.QToolButton] = _btn_ins + _btn_rem
        self._separator_label = QtW.QLabel()
        self._separator: str | None = None

        empty = QtW.QWidget()
        empty.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Preferred
        )
        layout.addWidget(empty)  # empty space
        layout.addWidget(self._info_label)
        layout.addWidget(self._separator_label)
        layout.addWidget(groupbox_ins)
        layout.addWidget(groupbox_rem)
        layout.addWidget(QSelectionRangeEdit(table))

    def update_for_table(self, table: QSpreadsheet):
        shape = table.model()._arr.shape
        self._info_label.setText(f"Shape: {shape!r}")

    def update_theme(self, theme: Theme):
        """Update the theme of the control."""
        color = theme.foreground
        for btn in self._tool_buttons:
            icon_path = btn.property("icon_path")
            btn.setIcon(QColoredSVGIcon.fromfile(icon_path, color))


def _sl(idx: int, axis: Literal[0, 1]) -> tuple:
    if axis == 0:
        return idx, slice(None)
    else:
        return slice(None), idx


def _tool_btn(callback, icon: str) -> QtW.QToolButton:
    """Create a tool button with the given icon and callback."""
    btn = QtW.QToolButton()
    icon_path = ICON_PATH / f"{icon}.svg"
    btn.setProperty("icon_path", icon_path)
    btn.setToolTip(callback.__doc__)
    btn.clicked.connect(callback)
    return btn


class QToolButtonGroup(QtW.QGroupBox):
    """A group of tool buttons with a horizontal layout."""

    def __init__(self, parent: QtW.QWidget | None = None):
        super().__init__(parent)
        self.setStyleSheet("QToolButtonGroup {margin: 0px;}")
        self.setLayout(QtW.QHBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)

        inner = QtW.QWidget()
        inner_layout = QtW.QHBoxLayout(inner)
        self.setContentsMargins(0, 0, 0, 0)
        inner_layout.setContentsMargins(0, 0, 0, 0)
        inner_layout.setSpacing(1)
        self._inner_layout = inner_layout
        self.layout().addWidget(inner)

    def add_widgets(self, *widgets: QtW.QWidget):
        """Add a widget to the group."""
        for widget in widgets:
            self._inner_layout.addWidget(widget)
        return widgets


ORD_A = ord("A")
CHARS = list("ABCDEFGHIJKLMNOPQRSTUVWXYZ") + [""]
LONGEST = CHARS[:-1]


def _iter_char(start: int, stop: int):
    if stop >= 26**4:
        raise ValueError("Stop must be less than 26**4 - 1")
    base_repr = np.base_repr(start, 26)
    current = np.zeros(4, dtype=np.int8)
    offset = 4 - len(base_repr)
    for i, d in enumerate(base_repr):
        current[i + offset] = int(d, 26)

    current[:3] -= 1
    for _ in range(start, stop):
        yield "".join(CHARS[s] for s in current)
        current[3] += 1
        for i in [3, 2, 1]:
            if current[i] >= 26:
                over = current[i] - 25
                current[i] = 0
                current[i - 1] += over


def char_arange(start: int, stop: int | None = None):
    """A char version of np.arange.

    Examples
    --------
    ``` python
    char_arange(3)  # array(["A", "B", "C"])
    char_arange(25, 28)  # array(["Z", "AA", "AB"])
    ```
    """
    global LONGEST
    if stop is None:
        start, stop = 0, start
    nmax = len(LONGEST)
    if stop <= nmax:
        return np.array(LONGEST[start:stop], dtype="<U4")
    LONGEST = np.append(LONGEST, np.fromiter(_iter_char(nmax, stop), dtype="<U4"))
    return LONGEST[start:].copy()


def _dict_to_array(value: dict[str, str]) -> np.ndarray:
    keys = list(value.keys())
    values = list(value.values())
    max_column_length = max(len(k) for k in values)
    arr = np.zeros((max_column_length + 1, len(keys)), dtype=np.dtypes.StringDType())
    arr[0, :] = keys
    for i, column in enumerate(values):
        arr[1:, i] = column
    return arr


def _array_like_to_array(value) -> np.ndarray:
    table = np.asarray(value, dtype=np.dtypes.StringDType())
    if table.ndim < 2:
        table = table.reshape(-1, 1)
    return table


def _size_to_shrink(proj: NDArray[np.intp]) -> int:
    """Determine the number of rows/columns array should be shrunk to.

    >>> _num_to_shrink(np.array([4, 3, 0, 0]))  # 2
    >>> _num_to_shrink(np.array([4, 3, 6, 0]))  # 3
    >>> _num_to_shrink(np.array([4, 3]))  # 2
    """
    _len = proj.size
    for i in range(_len - 1, -1, -1):
        if proj[i] != 0:
            return i + 1
    return _len


@dataclass
class SpreadsheetConfigs:
    default_cell_width: int = config_field(
        default=75,
        tooltip="Default width (pixel) of cells.",
    )
    default_cell_height: int = config_field(
        default=22,
        tooltip="Default height (pixel) of cells.",
    )
    separator_on_copy: str = config_field(
        default="\\t",
        tooltip="Separator used when the content of table is copied to the clipboard.",
    )



from __future__ import annotations

from abc import ABC, abstractmethod
from contextlib import contextmanager
from functools import cache
import weakref
import numpy as np
from typing import TYPE_CHECKING, Any, Callable, Iterable, Iterator, MutableSequence
from himena.types import Size, WindowRect, Margins
from himena import anchor as _anc
from himena.utils.misc import iter_subclasses


if TYPE_CHECKING:
    from typing import Self
    from himena.widgets import BackendMainWindow, MainWindow


class Layout(ABC):
    def __init__(self, main: BackendMainWindow | None = None):
        self._anchor = _anc.NoAnchor
        if main:
            self._main_window_ref = weakref.ref(main)
        else:
            self._main_window_ref = _no_ref
        self._parent_layout_ref: Callable[[], LayoutContainer | None] = _no_ref

    @property
    @abstractmethod
    def rect(self) -> WindowRect:
        """Position and size of the sub-window."""

    @rect.setter
    def rect(self, value: tuple[int, int, int, int] | WindowRect) -> None: ...

    @abstractmethod
    def _serialize_layout(self) -> Any:
        """Serialize the layout instance."""

    @classmethod
    @abstractmethod
    def _deserialize_layout(cls, obj, main: MainWindow) -> Any:
        """Deserialize the layout instance."""

    @property
    def size(self) -> Size[int]:
        """Size of the object."""
        return self.rect.size()

    @size.setter
    def size(self, value: tuple[int, int]) -> None:
        self.rect = (self.rect.left, self.rect.top, *value)
        return None

    @property
    def anchor(self) -> _anc.WindowAnchor:
        return self._anchor

    @anchor.setter
    def anchor(self, anchor: _anc.WindowAnchor | None):
        if anchor is None:
            anchor = _anc.NoAnchor
        elif isinstance(anchor, str):
            anchor = self._anchor_from_str(anchor)
        elif not isinstance(anchor, _anc.WindowAnchor):
            raise TypeError(f"Expected WindowAnchor, got {type(anchor)}")
        self._anchor = anchor

    def _reanchor(self, size: Size):
        """Reanchor all windows if needed (such as minimized windows)."""
        if rect := self._anchor.apply_anchor(size, self.size):
            self.rect = rect

    def _anchor_from_str(self, anchor: str):
        rect = self.rect
        main = self._main_window_ref()
        if main is None:
            w0, h0 = 100, 100
        else:
            w0, h0 = main._area_size()
        if anchor in ("top-left", "top left", "top_left"):
            return _anc.TopLeftConstAnchor(rect.left, rect.top)
        elif anchor in ("top-right", "top right", "top_right"):
            return _anc.TopRightConstAnchor(w0 - rect.right, rect.top)
        elif anchor in ("bottom-left", "bottom left", "bottom_left"):
            return _anc.BottomLeftConstAnchor(rect.left, h0 - rect.bottom)
        elif anchor in ("bottom-right", "bottom right", "bottom_right"):
            return _anc.BottomRightConstAnchor(w0 - rect.right, h0 - rect.bottom)
        else:
            raise ValueError(f"Unknown anchor: {anchor}")


@cache
def get_layout_class(typ: str) -> type[Layout]:
    for ly in iter_subclasses(Layout):
        if ly.__name__.lower() == typ:
            return ly
    raise ValueError(f"Unknown layout type: {typ}")


def construct_layout(obj, main: MainWindow) -> Layout:
    cls = get_layout_class(obj["type"])
    return cls._deserialize_layout(obj, main)


def _no_ref() -> None:
    return None


class EmptyLayout(Layout):
    """A layout that does not contain anything."""

    def __init__(self, main: BackendMainWindow | None = None):
        super().__init__(main)
        self._rect = WindowRect(0, 0, 0, 0)

    @property
    def rect(self):
        return self._rect

    @rect.setter
    def rect(self, value):
        rect_old = self._rect
        self._rect = WindowRect.from_tuple(*value)
        if parent := self._parent_layout_ref():
            parent._adjust_child_resize(self, rect_old, self._rect)

    def _serialize_layout(self):
        return {"type": "empty", "rect": self.rect}

    @classmethod
    def _deserialize_layout(cls, obj, main: MainWindow):
        self = cls(main._backend_main_window)
        self.rect = obj["rect"]
        return self


class LayoutContainer(Layout):
    """Layout that can contain other layouts."""

    def __init__(self, main: BackendMainWindow | None = None):
        self._rect = WindowRect(0, 0, 1000, 1000)
        super().__init__(main)
        self._anchor = _anc.AllCornersAnchor()
        self._is_calling_adjust_child_resize = False

    @contextmanager
    def _adjust_child_resize_context(self):
        was = self._is_calling_adjust_child_resize
        self._is_calling_adjust_child_resize = True
        try:
            yield
        finally:
            self._is_calling_adjust_child_resize = was

    @property
    def rect(self):
        return self._rect

    @rect.setter
    def rect(self, value):
        rect = WindowRect.from_tuple(*value)
        rect_old = self._rect
        self._rect = rect
        self._resize_children(rect)
        if parent := self._parent_layout_ref():
            parent._adjust_child_resize(self, rect_old, rect)

    @abstractmethod
    def _resize_children(self, rect: WindowRect):
        """Resize all children layouts based on the geometry of this layout."""

    @abstractmethod
    def remove(self, child: Layout) -> None:
        """Remove a child layout from this layout."""

    # @abstractmethod
    def _adjust_child_resize(
        self, child: Layout, rect_old: WindowRect, rect_new: WindowRect
    ):
        """Adjust layout container based on the child resize/move."""


class Layout1D(LayoutContainer, MutableSequence[Layout]):
    """Layout container that arranges children in 1D at the constant interval.

    Properties `margins` and `spacing` are defined as follows.
    ```
            spacing                 margin
             > <                     > <
    [ [child1] [   child2   ] [child3] ]
               <-  stretch ->
    ```

    Abstract methods:
    - `_resize_children(self, rect: WindowRect) -> None`
    - `insert(self, index: int, child: Layout) -> None`
    """

    def __init__(
        self,
        main: BackendMainWindow | None = None,
        *,
        margins: Margins[int] | tuple[int, int, int, int] = (0, 0, 0, 0),
        spacing: int = 0,
    ):
        super().__init__(main)
        self._children: list[Layout] = []
        self._margins = Margins(*margins)
        self._spacing = spacing

    @property
    def margins(self) -> Margins[int]:
        """Margins around the layout."""
        return self._margins

    @margins.setter
    def margins(self, value: Margins[int] | tuple[int, int, int, int]):
        self._margins = Margins(*value)
        self._resize_children(self.rect)

    @property
    def spacing(self) -> int:
        """Spacing between children."""
        return self._spacing

    @spacing.setter
    def spacing(self, value: int):
        if value < 0 or not isinstance(value, (int, np.int_)):
            raise ValueError(f"spacing must be non-negative integer, got {value}")
        self._spacing = value
        self._resize_children(self.rect)

    def set_margins(
        self,
        *,
        left: int | None = None,
        top: int | None = None,
        right: int | None = None,
        bottom: int | None = None,
    ):
        """Update margins around the layout."""
        margins_old = self.margins
        left = left if left is not None else margins_old.left
        top = top if top is not None else margins_old.top
        right = right if right is not None else margins_old.right
        bottom = bottom if bottom is not None else margins_old.bottom
        self.margins = left, top, right, bottom
        return None

    def __len__(self) -> int:
        return len(self._children)

    def __iter__(self) -> Iterator[Layout]:
        for _, child in self._children:
            yield child

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._children!r})"

    def __getitem__(self, key) -> Layout:
        _assert_supports_index(key)
        return self._children[key]

    def __setitem__(self, key, layout: Layout):
        if not isinstance(layout, Layout):
            raise TypeError(f"Can only set a Layout object, got {type(layout)}")
        _assert_supports_index(key)
        self._children[key] = layout
        layout._main_window_ref = self._main_window_ref
        layout._parent_layout_ref = weakref.ref(self)
        self._resize_children(self.rect)

    def remove(self, child: Layout) -> None:
        MutableSequence.remove(self, child)
        child._parent_layout_ref = _no_ref
        if len(self) > 1:
            self._resize_children(self.rect)
        elif len(self) == 1:
            self.remove(self[0])

    def _serialize_layout(self):
        return {
            "type": type(self).__name__.lower(),
            "children": [child._serialize_layout() for child in self._children],
            "margins": list(self.margins),
            "spacing": self.spacing,
        }

    @classmethod
    def _deserialize_layout(cls, obj, main: MainWindow) -> Self:
        self = cls(
            main._backend_main_window, margins=obj["margins"], spacing=obj["spacing"]
        )
        for child_obj in obj["children"]:
            child = construct_layout(child_obj, main)
            self.append(child)
        return self


class BoxLayout1D(Layout1D):
    def __init__(self, main=None, *, margins=(0, 0, 0, 0), spacing=0):
        super().__init__(main, margins=margins, spacing=spacing)
        self._stretches: list[float] = []

    def insert(self, index: int, child: Layout, *, stretch: float = 1) -> None:
        """Insert a child layout at the specified index."""
        if stretch <= 0:
            raise ValueError(f"stretch must be positive, got {stretch!r}")
        if not isinstance(child, Layout):
            raise TypeError(f"Can only insert a Layout object, got {type(child)}")
        self._children.insert(index, child)
        child._main_window_ref = self._main_window_ref
        child._parent_layout_ref = weakref.ref(self)
        self._stretches.insert(index, float(stretch))
        self._resize_children(self.rect)
        return self

    def append(self, child: Layout, *, stretch: float = 1) -> None:
        return self.insert(len(self), child, stretch=stretch)

    def add(
        self,
        child: Layout | Iterable[Layout],
        *more: Layout,
        stretch: float = 1,
    ) -> Self:
        """Add child layout(s) to the layout."""
        if not isinstance(child, Layout):
            child, more = child[0], [*child[1:], *more]
        self.append(child, stretch=stretch)
        for ch in more:
            self.append(ch, stretch=stretch)
        return self

    def __delitem__(self, key: int):
        _assert_supports_index(key)
        child = self._children.pop(key)
        child._parent_layout_ref = _no_ref
        del self._stretches[key]
        self._resize_children(self.rect)

    def add_vbox_layout(self, *, margins=(0, 0, 0, 0), spacing=0) -> VBoxLayout:
        layout = VBoxLayout(self._main_window_ref(), margins=margins, spacing=spacing)
        self.append(layout)
        return layout

    def add_hbox_layout(self, *, margins=(0, 0, 0, 0), spacing=0) -> HBoxLayout:
        layout = HBoxLayout(self._main_window_ref(), margins=margins, spacing=spacing)
        self.append(layout)
        return layout

    def _may_take_child(
        self, child: Layout, rect_old: WindowRect, rect_new: WindowRect
    ) -> Layout:
        # window moved
        dist2 = (rect_old.top - rect_new.top) ** 2 + (
            rect_old.left - rect_new.left
        ) ** 2
        if dist2 > 60**2:
            # remove window from the layout
            self.remove(child)
        self._resize_children(self.rect)


def _assert_supports_index(key):
    if not hasattr(key, "__index__"):
        raise TypeError(f"{key!r} cannot be used as an index")


class VBoxLayout(BoxLayout1D):
    """A vertical box layout."""

    def _iter_edge_and_span(self, rect: WindowRect) -> Iterator[tuple[int, int]]:
        num = len(self._children)
        if num == 0:
            yield from ()
            return
        h_cumsum = np.cumsum([0] + self._stretches, dtype=np.float32)
        edges = (h_cumsum / h_cumsum[-1] * rect.height).astype(np.int32)
        dy = self.spacing // 2
        edges[0] += self._margins.top - dy
        edges[-1] += self._margins.bottom + dy
        for i in range(num):
            top = edges[i] + dy
            height = edges[i + 1] - edges[i] - self.spacing
            yield top, height

    def _ortho_region(self, rect: WindowRect) -> tuple[int, int]:
        width = rect.width - self._margins.left - self._margins.right
        left = rect.left + self._margins.left
        return left, width

    def _resize_children(self, rect: WindowRect):
        left, width = self._ortho_region(rect)
        with self._adjust_child_resize_context():
            for i, (top, height) in enumerate(self._iter_edge_and_span(rect)):
                irect = WindowRect(left, top, width, height)
                self._children[i].rect = irect
        return None

    def _adjust_child_resize(self, child: Layout, rect_old, rect_new):
        if self._is_calling_adjust_child_resize:
            return
        top_changed = rect_old.top != rect_new.top
        bottom_changed = rect_old.bottom != rect_new.bottom
        with self._adjust_child_resize_context():
            if top_changed and bottom_changed:
                return self._may_take_child(child, rect_old, rect_new)

            top, height = self._ortho_region(self.rect)
            new_rect = child.rect
            stretches = self._stretches.copy()
            for i, (left, width) in enumerate(self._iter_edge_and_span(self.rect)):
                stretches[i] = width
                if top_changed and self._children[i] is child:
                    if i == 0:
                        child.rect = WindowRect(left, top, width, height)
                    else:
                        old_sum = stretches[i - 1] + stretches[i]
                        stretches[i - 1] = old_sum - new_rect.height
                        stretches[i] = new_rect.height
                elif bottom_changed and self._children[i - 1] is child:
                    if i == len(self) - 1:
                        child.rect = WindowRect(left, top, width, height)
                    else:
                        old_sum = stretches[i - 1] + stretches[i]
                        stretches[i] = old_sum - new_rect.height
                        stretches[i - 1] = new_rect.height
                else:
                    child.rect = WindowRect(left, top, width, height)
            self._stretches = stretches
            self._resize_children(self.rect)


class HBoxLayout(BoxLayout1D):
    """A horizontal box layout."""

    def _iter_edge_and_span(self, rect: WindowRect) -> Iterator[tuple[int, int]]:
        num = len(self._children)
        if num == 0:
            yield from ()
            return
        w_cumsum = np.cumsum([0] + self._stretches, dtype=np.float32)
        edges = (w_cumsum / w_cumsum[-1] * rect.width).astype(np.int32)
        dx = self.spacing // 2
        edges[0] += self._margins.left - dx
        edges[-1] += self._margins.right + dx
        for i in range(num):
            left = edges[i] + dx
            width = edges[i + 1] - edges[i] - self.spacing
            yield left, width

    def _ortho_region(self, rect: WindowRect) -> tuple[int, int]:
        height = rect.height - self._margins.top - self._margins.bottom
        top = rect.top + self._margins.top
        return top, height

    def _resize_children(self, rect: WindowRect):
        top, height = self._ortho_region(rect)
        with self._adjust_child_resize_context():
            for i, (left, width) in enumerate(self._iter_edge_and_span(rect)):
                irect = WindowRect(left, top, width, height)
                self._children[i].rect = irect
        return None

    def _adjust_child_resize(self, child: Layout, rect_old, rect_new):
        if self._is_calling_adjust_child_resize:
            return
        left_changed = rect_old.left != rect_new.left
        right_changed = rect_old.right != rect_new.right
        with self._adjust_child_resize_context():
            if left_changed and right_changed:
                return self._may_take_child(child, rect_old, rect_new)

            top, height = self._ortho_region(self.rect)
            new_rect = child.rect
            stretches = self._stretches.copy()
            for i, (left, width) in enumerate(self._iter_edge_and_span(self.rect)):
                stretches[i] = width
                if left_changed and self._children[i] is child:
                    if i == 0:
                        child.rect = WindowRect(left, top, width, height)
                    else:
                        old_sum = stretches[i - 1] + stretches[i]
                        w0 = stretches[i - 1] = old_sum - new_rect.width
                        self._children[i - 1].rect = self._children[
                            i - 1
                        ].rect.with_width(w0)
                        stretches[i] = new_rect.width
                elif right_changed and self._children[i - 1] is child:
                    if i == len(self) - 1:
                        child.rect = WindowRect(left, top, width, height)
                    else:
                        old_sum = stretches[i - 1] + stretches[i]
                        w0 = stretches[i] = old_sum - new_rect.width
                        self._children[i].rect = new_rect.with_width(w0)
                        stretches[i - 1] = new_rect.width
                else:
                    child.rect = WindowRect(left, top, width, height)
            self._stretches = stretches
            self._resize_children(self.rect)


class GridLayout(LayoutContainer):
    def __init__(
        self,
        main: BackendMainWindow | None = None,
        *,
        margins: Margins[int] | tuple[int, int, int, int] = (0, 0, 0, 0),
        spacing: Size[int] | tuple[int, int] = 0,
    ):
        super().__init__(main)
        self._margins = Margins(*margins)
        self._spacing = Size(*spacing)

    @property
    def margins(self) -> Margins[int]:
        """Margins around the layout."""
        return self._margins

    @margins.setter
    def margins(self, value: Margins[int] | tuple[int, int, int, int]):
        self._margins = Margins(*value)
        self._resize_children(self.rect)

    @property
    def spacing(self) -> Size:
        """Spacing between children."""
        return self._spacing

    @spacing.setter
    def spacing(self, value: Size[int] | tuple[int, int]):
        self._spacing = Size(*value)
        self._resize_children(self.rect)


class VStackLayout(Layout1D):
    def __init__(
        self,
        main: BackendMainWindow | None = None,
        *,
        margins: Margins[int] | tuple[int, int, int, int] = (0, 0, 0, 0),
        spacing: int = 0,
        inverted: bool = False,
    ):
        super().__init__(main, margins=margins, spacing=spacing)
        self._inverted = inverted

    @property
    def inverted(self) -> bool:
        return self._inverted

    @inverted.setter
    def inverted(self, value: bool):
        self._inverted = bool(value)
        self._resize_children(self.rect)

    def insert(self, index: int, child: Layout) -> None:
        """Insert a child layout at the specified index."""
        self._children.insert(index, child)
        child._main_window_ref = self._main_window_ref
        child._parent_layout_ref = weakref.ref(self)
        self._resize_children(self.rect)
        return self

    def append(self, child: Layout) -> None:
        return self.insert(len(self), child)

    def add(self, child: Layout) -> Self:
        self.append(child)
        return self

    def __delitem__(self, key: int):
        _assert_supports_index(key)
        self._children[key]._parent_layout_ref = _no_ref
        del self._children[key]
        self._resize_children(self.rect)

    def _resize_children(self, rect):
        num = len(self._children)
        if num == 0:
            return
        heights = [ch.rect.height for ch in self._children]
        h_cumsum = np.cumsum([0] + heights, dtype=np.uint32)
        if self._inverted:
            bottoms = rect.bottom - h_cumsum[:-1] + self._margins.bottom
            for i, child in enumerate(self._children):
                child.rect = child.rect.move_bottom_left(rect.left, bottoms[i])
        else:
            tops = h_cumsum[:-1] + rect.top + self._margins.top
            for i, child in enumerate(self._children):
                child.rect = child.rect.move_top_left(rect.left, tops[i])



from __future__ import annotations
from concurrent.futures import Future, ThreadPoolExecutor
from typing import Any

from psygnal import Signal
from app_model.registries import CommandsRegistry as _CommandsRegistry

from himena.exceptions import Cancelled


class CommandsRegistry(_CommandsRegistry):
    """A command registry that emits signal when command is executed."""

    executed = Signal(str, Future)  # id

    def execute_command(
        self,
        id: str,
        *args: Any,
        execute_asynchronously: bool = False,
        **kwargs: Any,
    ) -> Future:
        try:
            cmd = self[id].run_injected
        except KeyError as e:
            raise KeyError(f"Command {id!r} not registered") from e  # pragma: no cover

        if execute_asynchronously:
            with ThreadPoolExecutor() as executor:
                return executor.submit(cmd, *args, **kwargs)

        future: Future = Future()
        try:
            future.set_result(cmd(*args, **kwargs))
        except Cancelled as exc:
            future.set_result(exc)
            return future  # do not record
        except Exception as e:
            if self._raise_synchronous_exceptions:
                # note, the caller of this function can also achieve this by
                # calling `future.result()` on the returned future object.
                raise e
            future.set_exception(e)

        self.executed.emit(id, future)
        return future



from __future__ import annotations

from typing import TYPE_CHECKING
import weakref
from qtpy import QtCore, QtWidgets as QtW, QtGui
from app_model.types import Action
from himena.plugins import validate_protocol, update_config_context
from himena_builtins.qt.favorites._config import FavoriteCommandsConfig

if TYPE_CHECKING:
    from himena.widgets import MainWindow


class QFavoriteCommands(QtW.QWidget):
    """List of favorite commands.

    You can add commands to this list from plugin configuration in the setting dialog,
    or directly drag-and-drop commands from the Command History dock widget.
    """

    def __init__(self, ui: MainWindow):
        super().__init__()
        self._ui_ref = weakref.ref(ui)
        self._command_list = QCommandList(self)
        layout = QtW.QVBoxLayout(self)

        layout.addWidget(self._command_list)

    @validate_protocol
    def update_configs(self, cfg: FavoriteCommandsConfig) -> None:
        if not (ui := self._ui_ref()):
            return
        self._command_list.clear()
        for cmd_id in cfg.commands:
            if action := ui.model_app.registered_actions.get(cmd_id):
                self._command_list.add_action(action)


_ID_ROLE = QtCore.Qt.ItemDataRole.UserRole


class QCommandList(QtW.QListWidget):
    def __init__(self, parent: QFavoriteCommands):
        super().__init__(parent)
        self._ui_ref = parent._ui_ref
        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.setDefaultDropAction(QtCore.Qt.DropAction.MoveAction)
        self.setSelectionMode(QtW.QAbstractItemView.SelectionMode.SingleSelection)
        self.setDragDropMode(QtW.QAbstractItemView.DragDropMode.InternalMove)

    def add_action(self, action: Action):
        btn = QCommandPushButton(action.title)
        btn.clicked.connect(_make_callback(self._ui_ref(), action.id))
        tooltip = f"{action.tooltip}\n(command ID: {action.id})"
        btn.setToolTip(tooltip)
        item = QtW.QListWidgetItem(self)
        item.setSizeHint(btn.sizeHint())
        item.setData(_ID_ROLE, action.id)
        self.setItemWidget(item, btn)
        btn.delete_requested.connect(lambda: self.delete_item(item))

    def has_action(self, action_id: str) -> bool:
        for i in range(self.count()):
            item = self.item(i)
            if item.data(_ID_ROLE) == action_id:
                return True
        return False

    def delete_item(self, item: QtW.QListWidgetItem):
        """Delete the item from the list."""
        self.takeItem(self.row(item))
        self._update_plugin_config()

    def _update_plugin_config(self):
        with update_config_context(
            config_class=FavoriteCommandsConfig,
            plugin_id="builtins:favorite-commands",
            update_widget=False,
        ) as cfg:
            cfg.commands = [self.item(i).data(_ID_ROLE) for i in range(self.count())]

    def dragEnterEvent(self, e):
        e.accept()
        return super().dragEnterEvent(e)

    def dragMoveEvent(self, e):
        e.accept()
        return super().dragMoveEvent(e)

    def dropEvent(self, e: QtGui.QDropEvent):
        if e.source() is self:
            return super().dropEvent(e)
        self._drop_mime_data(e.mimeData())

    def _drop_mime_data(self, mime: QtCore.QMimeData) -> None:
        """Handle the drop event."""
        if not (ui := self._ui_ref()):
            return
        if mime.hasFormat("text/command-id"):
            command_id = bytes(mime.data("text/command-id")).decode("utf-8")
            if not self.has_action(command_id):
                if action := ui.model_app.registered_actions.get(command_id):
                    self.add_action(action)
                    self._update_plugin_config()


class QCommandPushButton(QtW.QPushButton):
    delete_requested = QtCore.Signal(object)

    def __init__(self, command: str, parent: QtW.QWidget | None = None):
        super().__init__(command, parent)
        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self._show_context_menu)
        self._command_id = command

    def _make_context_menu(self) -> QtW.QMenu:
        """Create the context menu."""
        menu = QtW.QMenu(self)
        action = menu.addAction("Delete")
        action.triggered.connect(lambda: self.delete_requested.emit(self))
        return menu

    def _show_context_menu(self, pos: QtCore.QPoint):
        """Show the context menu."""
        self._make_context_menu().exec(self.mapToGlobal(pos))


def _make_callback(ui: MainWindow, cmd: str):
    """Create a callback for the command."""

    def callback():
        ui.exec_action(cmd)

    return callback



from pathlib import Path
from magicgui.type_map import TypeMap
from himena.qt.magicgui._basic_widgets import (
    IntEdit,
    FloatEdit,
    IntListEdit,
    FloatListEdit,
)
from himena.qt.magicgui._modeldrop import (
    ModelDrop,
    ModelListDrop,
    SubWindowDrop,
    SubWindowListDrop,
)
from himena.qt.magicgui._toggle_switch import ToggleSwitch
from himena.qt.magicgui._color import ColorEdit, ColormapEdit
from himena.qt.magicgui._file_edit import FileEdit
from himena.types import WidgetDataModel
from himena.widgets import SubWindow
from cmap import Color, Colormap

TYPE_MAP = TypeMap()


def register_magicgui_types():
    """Register magicgui types."""

    TYPE_MAP.register_type(WidgetDataModel, widget_type=ModelDrop)
    TYPE_MAP.register_type(list[WidgetDataModel], widget_type=ModelListDrop)
    TYPE_MAP.register_type(SubWindow, widget_type=SubWindowDrop)
    TYPE_MAP.register_type(list[SubWindow], widget_type=SubWindowListDrop)
    TYPE_MAP.register_type(bool, widget_type=ToggleSwitch)
    TYPE_MAP.register_type(int, widget_type=IntEdit)
    TYPE_MAP.register_type(float, widget_type=FloatEdit)
    TYPE_MAP.register_type(list[int], widget_type=IntListEdit)
    TYPE_MAP.register_type(list[float], widget_type=FloatListEdit)
    TYPE_MAP.register_type(Color, widget_type=ColorEdit)
    TYPE_MAP.register_type(Colormap, widget_type=ColormapEdit)
    TYPE_MAP.register_type(Path, widget_type=FileEdit)

    # remove non-serializable types from the magicgui type map
    TYPE_MAP._simple_types.pop(range, None)
    TYPE_MAP._simple_types.pop(slice, None)


def get_type_map():
    """Get the magicgui type map for himena."""
    return TYPE_MAP



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Mapping
import weakref

from cmap import Color, Colormap
import numpy as np
from qtpy import QtGui, QtCore, QtWidgets as QtW
from qtpy.QtCore import Qt

from himena import MainWindow
from himena.consts import StandardType
from himena.types import Size, WidgetDataModel, Parametric
from himena.standards import plotting as hplt, roi as _roi
from himena.standards.model_meta import DataFrameMeta, TableMeta, DataFramePlotMeta
from himena.utils.collections import UndoRedoStack
from himena.plugins import validate_protocol, register_function, config_field
from himena.qt import drag_command
from himena.data_wrappers import wrap_dataframe, DataFrameWrapper
from himena_builtins.qt.widgets._table_components import (
    QTableBase,
    QSelectionRangeEdit,
    format_table_value,
    QHorizontalHeaderView,
    Editability,
    FLAGS,
    parse_string,
)
from himena_builtins.qt.widgets._splitter import QSplitterHandle
from himena_builtins.qt.widgets._dragarea import QDraggableArea

if TYPE_CHECKING:
    from himena_builtins.qt.widgets._table_components._selection_model import Index


class QDataFrameModel(QtCore.QAbstractTableModel):
    """Table model for data frame."""

    def __init__(self, df: DataFrameWrapper, transpose: bool = False, parent=None):
        super().__init__(parent)
        self._df = df
        self._transpose = transpose
        self._cfg = DataFrameConfigs()
        # this set records all the updated column indices for copy-on-write behavior
        self._updated_columns: set[int] = set()

    @property
    def df(self) -> DataFrameWrapper:
        return self._df

    def rowCount(self, parent=None):
        if self._transpose:
            return self.df.num_columns()
        return self.df.num_rows()

    def columnCount(self, parent=None):
        if self._transpose:
            return self.df.num_rows()
        return self.df.num_columns()

    def data(
        self,
        index: QtCore.QModelIndex,
        role: Qt.ItemDataRole = Qt.ItemDataRole.DisplayRole,
    ):
        if self._transpose:
            r, c = index.column(), index.row()
        else:
            r, c = index.row(), index.column()
        if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
            df = self.df
            if r < df.num_rows() and c < df.num_columns():
                value = df[r, c]
                dtype = df.get_dtype(c)
                if role == Qt.ItemDataRole.DisplayRole:
                    text = format_table_value(value, dtype.kind)
                else:
                    text = str(value)
                return text
        return QtCore.QVariant()

    def setData(self, index: QtCore.QModelIndex, value: Any, role: int = ...) -> bool:
        if role == Qt.ItemDataRole.EditRole:
            i_row, i_col = index.row(), index.column()
            val = parse_string(value, self.df.get_dtype(i_col).kind)
            name = self.df.column_names()[i_col]
            col = self.df.column_to_array(name)
            if i_col not in self._updated_columns:
                col = col.copy()
                self._updated_columns.add(i_col)
            col[i_row] = val
            self._df = self.df.with_columns({name: col})
            return True
        return False

    def flags(self, index):
        return FLAGS

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ):
        if self._transpose:
            is_header = orientation == Qt.Orientation.Vertical
        else:
            is_header = orientation == Qt.Orientation.Horizontal
        if is_header:
            if role == Qt.ItemDataRole.DisplayRole:
                if section >= self.df.num_columns():
                    return None
                return str(self.df.column_names()[section])
            elif role == Qt.ItemDataRole.ToolTipRole:
                if section < self.df.num_columns():
                    return self._column_tooltip(section)
                return None

        else:
            if role == Qt.ItemDataRole.DisplayRole:
                return str(section)

    def _column_tooltip(self, section: int):
        name = self.df.column_names()[section]
        dtype = self.df.get_dtype(section)
        return f"{name} (dtype: {dtype.name})"


class QDraggableHorizontalHeader(QHorizontalHeaderView):
    """Header view for DataFrameView that supports drag and drop."""

    def __init__(self, parent: QDataFrameView):
        super().__init__(parent)
        self._table_view_ref = weakref.ref(parent)
        self.setMouseTracking(True)
        self._hover_drag_indicator = QDraggableArea(self)
        self._hover_drag_indicator.setWindowFlags(
            QtCore.Qt.WindowType.FramelessWindowHint
        )
        self._hover_drag_indicator.setFixedSize(14, 14)
        self._hover_drag_indicator.hide()
        self._hover_drag_indicator.dragged.connect(self._drag_event)
        self._drag_enabled = True

    def mouseMoveEvent(self, e: QtGui.QMouseEvent):
        view = self._table_view_ref()
        if view is None or not self._drag_enabled:
            return super().mouseMoveEvent(e)
        if e.button() == QtCore.Qt.MouseButton.NoButton:
            # hover
            index = self.logicalIndexAt(e.pos())
            self._process_move_event(index)
        return super().mouseMoveEvent(e)

    def _process_move_event(self, index: int):
        view = self._table_view_ref()
        hovered_column_selected = False
        for sel in view.selection_model.iter_col_selections():
            if sel.start <= index < sel.stop:
                hovered_column_selected = True
                break
        if hovered_column_selected and index >= 0:
            index_rect = self.visualRectAtIndex(index)
            top_right = index_rect.topRight()
            top_right.setX(top_right.x() - 14)
            dy = (index_rect.height() - self._hover_drag_indicator.height()) / 2
            top_right.setY(top_right.y() + int(dy))
            self._hover_drag_indicator.move(top_right)
            self._hover_drag_indicator.show()
        else:
            self._hover_drag_indicator.hide()

    def leaveEvent(self, a0):
        self._hover_drag_indicator.hide()
        return super().leaveEvent(a0)

    def _data_model_for_drag(self) -> QtGui.QDrag | None:
        view = self._table_view_ref()
        if view is None or not self._drag_enabled:
            return
        cols = [sel.start for sel in view.selection_model.iter_col_selections()]
        model = view.to_model()
        df = wrap_dataframe(model.value)
        s = "" if len(cols) == 1 else "s"
        return drag_command(
            view,
            "builtins:QDataFrameView:select-columns",
            StandardType.DATAFRAME,
            with_params={"columns": cols},
            desc=f"{df.num_columns()} column{s}",
            text_data=lambda: df.to_csv_string("\t"),
            exec=False,
        )

    def _drag_event(self):
        if drag := self._data_model_for_drag():
            drag.exec()


class QDataFrameView(QTableBase):
    """A table widget for viewing DataFrame.

    ## Basic Usage

    - This widget is a table widget for viewing a dataframe. Supported data types
      includes `dict[str, numpy.ndarray]`, `pandas.DataFrame`, `polars.DataFrame`,
      `pyarrow.Table` and `narwhals.DataFrame`.
    - `Ctrl+F` to search a string in the table.
    - Each item can be edited by double-clicking it, but only the standard scalar types
      are supported.

    ## Drag and Drop

    Selected columns can be dragged out as a model of type `StandardType.DATAFRAME`
    ("dataframe"). Use the drag indicator on the header to start dragging.
    """

    __himena_widget_id__ = "builtins:QDataFrameView"
    __himena_display_name__ = "Built-in DataFrame Viewer"

    def __init__(self, ui):
        super().__init__(ui)
        self._hor_header = QDraggableHorizontalHeader(self)
        self.setHorizontalHeader(self._hor_header)
        self.horizontalHeader().setFixedHeight(18)
        self.horizontalHeader().setDefaultSectionSize(75)
        self._control: QDataFrameViewControl | None = None  # deferred
        self._model_type = StandardType.DATAFRAME
        self._undo_stack = UndoRedoStack[EditAction](size=20)
        self._sep_on_copy = "\t"
        self._extension_default = ".csv"

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        df = wrap_dataframe(model.value)
        transpose = False
        if isinstance(meta := model.metadata, DataFrameMeta):
            transpose = meta.transpose
        self.setModel(QDataFrameModel(df, transpose=transpose))
        if df.num_rows() == 1 and transpose:
            # single-row, row-orinted table should be expanded
            self.resizeColumnsToContents()
        if ext := model.extension_default:
            self._extension_default = ext
        # update the table-widget-specific settings
        if isinstance(meta := model.metadata, TableMeta):
            self._selection_model.clear()
            if (pos := meta.current_position) is not None:
                index = self.model().index(*pos)
                self.setCurrentIndex(index)
                self._selection_model.current_index = pos
            for (r0, r1), (c0, c1) in meta.selections:
                self._selection_model.append((slice(r0, r1), slice(c0, c1)))

        self.control_widget().update_for_table(self)
        self._model_type = model.type
        self.update()
        return None

    @validate_protocol
    def to_model(self) -> WidgetDataModel:
        # NOTE: if this model is passed to another widget and modified in this widget,
        # the other dataframe will be modified as well. To avoid this, we need to reset
        # the copy-on-write state of this array.
        self.model()._updated_columns.clear()
        return WidgetDataModel(
            value=self.model().df.unwrap(),
            type=self.model_type(),
            extension_default=self._extension_default,
            metadata=self._prep_table_meta(cls=DataFrameMeta),
        )

    @validate_protocol
    def model_type(self) -> str:
        return self._model_type

    @validate_protocol
    def update_configs(self, cfg: DataFrameConfigs):
        self._sep_on_copy = cfg.separator_on_copy.encode().decode("unicode_escape")
        self._hor_header._drag_enabled = cfg.column_drag_enabled

    @validate_protocol
    def is_modified(self) -> bool:
        if self._modified_override is not None:
            return self._modified_override
        return self._undo_stack.undoable()

    @validate_protocol
    def set_modified(self, value: bool) -> None:
        self._modified_override = value

    @validate_protocol
    def is_editable(self) -> bool:
        return self.editTriggers() == Editability.TRUE

    @validate_protocol
    def set_editable(self, value: bool) -> None:
        if value:
            trig = Editability.TRUE
        else:
            trig = Editability.FALSE
        self.setEditTriggers(trig)

    @validate_protocol
    def control_widget(self):
        if self._control is None:
            self._control = QDataFrameViewControl()
        return self._control

    def keyPressEvent(self, e: QtGui.QKeyEvent) -> None:
        if e.matches(QtGui.QKeySequence.StandardKey.Copy):
            return self.copy_data()
        if (
            e.modifiers() & Qt.KeyboardModifier.ControlModifier
            and e.key() == QtCore.Qt.Key.Key_F
        ):
            self._find_string()
            return
        return super().keyPressEvent(e)

    def copy_data(self):
        rng = self._selection_model.get_single_range()

        rsl, csl = rng
        r0, r1 = rsl.start, rsl.stop
        c0, c1 = csl.start, csl.stop
        csv_text = self.model().df.get_subset(r0, r1, c0, c1).to_csv_string("\t")
        clipboard = QtGui.QGuiApplication.clipboard()
        clipboard.setText(csv_text)

    def edit_item(self, r: int, c: int, value: str):
        self.model().setData(self.model().index(r, c), value, Qt.ItemDataRole.EditRole)
        # datChanged needs to be emitted manually
        self.model().dataChanged.emit(
            self.model().index(r, c),
            self.model().index(r, c),
            [Qt.ItemDataRole.EditRole],
        )

    def _make_context_menu(self):
        menu = QtW.QMenu(self)
        menu.addAction("Copy", self.copy_data)
        return menu

    if TYPE_CHECKING:

        def model(self) -> QDataFrameModel: ...


@register_function(command_id="builtins:QDataFrameView:select-columns", menus=[])
def select_columns(model: WidgetDataModel) -> Parametric:
    """Select columns, used for dragging columns off a dataframe."""

    def run(columns: list[int]) -> WidgetDataModel:
        df = wrap_dataframe(model.value)
        nrows = df.num_rows()
        dict_out = {}
        for icol in columns:
            dict_out.update(df.get_subset(0, nrows, icol, icol + 1).to_dict())
        df = df.from_dict(dict_out)
        return model.with_value(df.unwrap())

    return run


class QDictView(QDataFrameView):
    """A widget for viewing dictionary with scalar values."""

    __himena_widget_id__ = "builtins:QDictView"
    __himena_display_name__ = "Built-in Dictionary Viewer"

    def __init__(self, ui: MainWindow):
        super().__init__(ui)
        self._extension_default = ".json"
        self._model_type = StandardType.DICT
        self.horizontalHeader().hide()

    @validate_protocol
    def update_model(self, model: WidgetDataModel[dict]):
        if not isinstance(model.value, Mapping):
            raise TypeError(f"Expected a mapping, got {type(model.value)}.")
        was_empty = self.model() is None
        df = wrap_dataframe({k: [v] for k, v in model.value.items()})
        self.setModel(QDataFrameModel(df, transpose=True))
        if was_empty:
            self.resizeColumnsToContents()
        if ext := model.extension_default:
            self._extension_default = ext
        self.control_widget().update_for_table(self)
        self._model_type = model.type
        self.update()
        return None

    @validate_protocol
    def to_model(self) -> WidgetDataModel:
        self.model()._updated_columns.clear()
        return WidgetDataModel(
            value={k: v[0] for k, v in self.model().df.to_dict().items()},
            type=self.model_type(),
            extension_default=self._extension_default,
        )

    @validate_protocol
    def size_hint(self):
        return 260, 260


class QDataFrameViewControl(QtW.QWidget):
    def __init__(self):
        super().__init__()
        _R_CENTER = (
            QtCore.Qt.AlignmentFlag.AlignRight | QtCore.Qt.AlignmentFlag.AlignVCenter
        )
        layout = QtW.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setAlignment(_R_CENTER)
        self._label = QtW.QLabel("")
        self._label.setAlignment(_R_CENTER)
        layout.addWidget(self._label)
        self._selection_range = QSelectionRangeEdit()
        layout.addWidget(self._selection_range)

    def update_for_table(self, table: QDataFrameView | None):
        if table is None:
            return
        model = table.model()
        self._label.setText(
            f"{model.df.type_name()} ({model.rowCount()}, {model.columnCount()})"
        )
        self._selection_range.connect_table(table)
        return None


class QDataFramePlotView(QtW.QSplitter):
    """A widget for viewing a dataframe on the left and its plot on the right.

    ## Basic Usage

    All the columns of the dataframe must be numerical data type. If there's only one
    column, it will be considered as the y values. If there are more, the first column
    will be the x values and the rest of the columns will be separate y values. If there
    are more than one set of y values, clicking the column will highlight the plot on
    the right.
    """

    __himena_widget_id__ = "builtins:QDataFramePlotView"
    __himena_display_name__ = "Built-in DataFrame Plot View"

    def __init__(self, ui: MainWindow):
        from himena_builtins.qt.plot._canvas import QModelMatplotlibCanvas

        super().__init__(QtCore.Qt.Orientation.Horizontal)
        self._table_widget = QDataFrameView(ui)
        self._table_widget.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Expanding
        )
        self._plot_widget = QModelMatplotlibCanvas()
        self._plot_widget.update_model(
            WidgetDataModel(value=hplt.figure(), type=StandardType.PLOT)
        )
        right = QtW.QWidget()
        right.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Expanding
        )
        layout_right = QtW.QVBoxLayout(right)
        layout_right.setContentsMargins(0, 0, 0, 0)
        layout_right.setSpacing(1)
        layout_right.addWidget(self._plot_widget._toolbar)
        layout_right.addWidget(self._plot_widget)
        self._model_type = StandardType.DATAFRAME_PLOT
        self._color_cycle: list[str] | None = None

        self.addWidget(self._table_widget)
        self.addWidget(right)
        self.setStretchFactor(0, 1)
        self.setStretchFactor(1, 2)

        self._table_widget.selection_model.moved.connect(
            self._update_plot_for_selections
        )
        self._y_column_names: list[str] = []

    def createHandle(self):
        return QSplitterHandle(self, side="left")

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        df = wrap_dataframe(model.value)
        col_names = df.column_names()
        if isinstance(meta := model.metadata, DataFramePlotMeta):
            plot_type = meta.plot_type
            plot_background_color = meta.plot_background_color
            plot_color_cycle_name = meta.plot_color_cycle
        else:
            plot_type = "line"
            plot_background_color = "#FFFFFF"
            plot_color_cycle_name = None
        if plot_color_cycle_name is None:
            if np.mean(Color(plot_background_color).rgba) > 0.5:
                plot_color_cycle = Colormap("tab10")
            else:
                plot_color_cycle = Colormap("colorbrewer:Dark2")
        else:
            plot_color_cycle = Colormap(plot_color_cycle_name)

        if len(col_names) == 0:
            raise ValueError("No columns in the dataframe.")
        elif len(col_names) == 1:
            x = np.arange(df.num_rows())
            self._y_column_names = col_names
        else:
            x = df.column_to_array(col_names[0])
            self._y_column_names = col_names[1:]
        fig = hplt.figure(background_color=plot_background_color)
        colors = plot_color_cycle.color_stops.colors
        if colors[0].rgba8[3] == 0:
            colors = colors[1:]
        for i, ylabel in enumerate(self._y_column_names):
            y = df.column_to_array(ylabel)
            color = colors[i % len(colors)]
            if plot_type == "line":
                fig.plot(x, y, color=color, name=ylabel)
            elif plot_type == "scatter":
                fig.scatter(x, y, color=color, name=ylabel)
            else:
                raise ValueError(f"Unsupported plot type: {plot_type!r}")
        self._table_widget.update_model(model)

        # update plot
        model_plot = WidgetDataModel(value=fig, type=StandardType.PLOT)
        self._plot_widget.update_model(model_plot)
        self._model_type = model.type
        self._color_cycle = [c.hex for c in colors]
        return None

    @validate_protocol
    def to_model(self) -> WidgetDataModel:
        meta = self._table_widget._prep_table_meta()
        return WidgetDataModel(
            value=self._table_widget.model().df.unwrap(),
            type=self.model_type(),
            extension_default=".csv",
            metadata=DataFramePlotMeta(
                current_position=meta.current_position,
                selections=meta.selections,
                plot_type="line",
                plot_color_cycle=self._color_cycle,
                plot_background_color="#FFFFFF",
                rois=_roi.RoiListModel(),
            ),
        )

    @validate_protocol
    def model_type(self) -> str:
        return self._model_type

    @validate_protocol
    def is_modified(self) -> bool:
        return self._table_widget.is_modified()

    @validate_protocol
    def control_widget(self):
        return self._table_widget.control_widget()

    @validate_protocol
    def size_hint(self):
        return 480, 300

    @validate_protocol
    def widget_added_callback(self):
        # adjuct size
        self.setSizes([160, self.width() - 160])
        self._plot_widget.widget_added_callback()
        return None

    @validate_protocol
    def widget_resized_callback(self, old: Size, new: Size):
        left_width = self._table_widget.width()
        old = old.with_width(max(old.width - left_width, 10))
        new = new.with_width(max(new.width - left_width, 10))
        self._plot_widget.widget_resized_callback(old, new)

    @validate_protocol
    def theme_changed_callback(self, theme):
        # self._table_widget.theme_changed_callback(theme)
        self._plot_widget.theme_changed_callback(theme)

    def _update_plot_for_selections(self, old: Index, new: Index):
        axes_layout = self._plot_widget._plot_models
        if not isinstance(axes_layout, hplt.SingleAxes):
            return
        inds = set()
        for sl in self._table_widget.selection_model.iter_col_selections():
            inds.update(range(sl.start, sl.stop))
        inds.discard(0)  # x axis
        if len(inds) == 0:
            inds = set(range(1, len(self._y_column_names) + 1))

        selected_names = [self._y_column_names[i - 1] for i in inds]

        for model in axes_layout.axes.models:
            selected = model.name in selected_names
            if isinstance(model, hplt.Line):
                model.edge.alpha = 1.0 if selected else 0.4
            elif isinstance(model, hplt.Scatter):
                model.face.alpha = 1.0 if selected else 0.4
                model.edge.alpha = 1.0 if selected else 0.4
        self._plot_widget.update_model(
            WidgetDataModel(value=axes_layout, type=StandardType.PLOT)
        )
        return None


@dataclass
class DataFrameConfigs:
    column_drag_enabled: bool = config_field(default=True)
    separator_on_copy: str = config_field(
        default="\\t",
        tooltip="Separator used when the content of table is copied to the clipboard.",
    )


@dataclass
class EditAction:
    old: Any
    new: Any
    index: Any

    def invert(self) -> EditAction:
        return EditAction(self.new, self.old, self.index)

    def apply(self, table: QDataFrameView):
        raise NotImplementedError



from app_model.types import (
    KeyBindingRule,
    KeyCode,
    KeyMod,
    StandardKeyBinding,
)
from himena.utils.collections import OrderedSet
from himena.consts import MenuId
from himena.exceptions import Cancelled
from himena.plugins._signature import configure_gui
from himena.widgets import MainWindow, SubWindow
from himena.types import (
    Parametric,
    WindowState,
)
from himena._app_model._context import AppContext as _ctx
from himena._app_model.actions._registry import ACTIONS, SUBMENUS

WINDOW_GROUP = "00_window"
LAYOUT_GROUP = "03_layout"


@ACTIONS.append_from_fn(
    id="new-tab",
    title="New Tab",
    menus=[MenuId.VIEW],
    keybindings=[StandardKeyBinding.AddTab],
)
def new_tab(ui: MainWindow) -> None:
    """Create a new tab."""
    ui.add_tab()


@ACTIONS.append_from_fn(
    id="close-tab",
    title="Close Tab",
    enablement=_ctx.num_tabs > 0,
    menus=[MenuId.VIEW],
    keybindings=[KeyBindingRule(primary=KeyMod.CtrlCmd | KeyMod.Shift | KeyCode.KeyW)],
)
def close_current_tab(ui: MainWindow) -> None:
    """Close the current tab."""
    idx = ui._backend_main_window._current_tab_index()
    if idx is None:
        return
    win_modified = [win for win in ui.tabs[idx] if win._need_ask_save_before_close()]
    if len(win_modified) > 0 and ui._instructions.confirm:
        _modified_msg = "\n".join([f"- {win.title}" for win in win_modified])
        ans = ui.exec_choose_one_dialog(
            "Close the tab?",
            f"Some windows in this tab are not saved yet:\n{_modified_msg}\nClose the "
            "tab without saving?",
            ["Yes, close all", "No"],
        )
        if ans == "No":
            raise Cancelled
    ui.tabs.pop(idx)


@ACTIONS.append_from_fn(
    id="goto-last-tab",
    title="Go To Last Tab",
    enablement=_ctx.num_tabs > 1,
    menus=[MenuId.VIEW],
    keybindings=[KeyBindingRule(primary=KeyMod.CtrlCmd | KeyCode.Tab)],
    recording=False,
)
def goto_last_tab(ui: MainWindow) -> None:
    """Go to the last tab."""
    if (idx := ui._history_tab.get_from_last(2)) is not None:
        ui.tabs.current_index = idx


@ACTIONS.append_from_fn(
    id="merge-tabs",
    title="Merge Tabs ...",
    enablement=_ctx.num_tabs > 1,
    menus=[MenuId.VIEW],
    need_function_callback=True,
)
def merge_tabs(ui: MainWindow) -> Parametric:
    """Select tabs and merge them."""
    if len(ui.tabs) < 2:
        return

    @configure_gui(names={"choices": ui.tabs.names, "widget_type": "Select"})
    def choose_tabs_to_merge(names: list[str]) -> None:
        if len(names) < 2:
            return
        i_tab = ui.tabs.names.index(names[0])
        for name in names[1:]:
            for window in ui.tabs[name]:
                ui.move_window(window, i_tab)
            del ui.tabs[name]
        return None

    return choose_tabs_to_merge


@ACTIONS.append_from_fn(
    id="minimize-other-windows",
    title="Minimize Other Windows",
    menus=[{"id": MenuId.VIEW, "group": WINDOW_GROUP}],
    enablement=(_ctx.num_sub_windows > 1) & _ctx.is_subwindow_focused,
)
def minimize_others(ui: MainWindow):
    """Minimize all sub-windows except the current one."""
    if area := ui.tabs.current():
        if cur_window := area.current():
            for window in area:
                if cur_window is window:
                    continue
                window.state = WindowState.MIN
            return
    raise RuntimeError("No window focused.")


@ACTIONS.append_from_fn(
    id="show-all-windows",
    title="Show All Windows",
    menus=[{"id": MenuId.VIEW, "group": WINDOW_GROUP}],
    enablement=_ctx.num_sub_windows > 0,
)
def show_all_windows(ui: MainWindow):
    """Show all sub-windows in the current tab."""
    if area := ui.tabs.current():
        for window in area:
            if window.state is WindowState.MIN:
                window.state = WindowState.NORMAL


@ACTIONS.append_from_fn(
    id="tile-windows",
    title="Tile Windows",
    enablement=_ctx.num_sub_windows > 1,
    menus=[{"id": MenuId.VIEW, "group": WINDOW_GROUP}],
)
def tile_windows(ui: MainWindow) -> None:
    """Tile all the windows."""
    if area := ui.tabs.current():
        area.tile_windows()


@ACTIONS.append_from_fn(
    id="collect-windows",
    title="Collect Windows From Other Tabs",
    menus=[{"id": MenuId.VIEW, "group": WINDOW_GROUP}],
    enablement=_ctx.num_tabs > 1,
    need_function_callback=True,
)
def collect_windows(ui: MainWindow) -> Parametric:
    """Collect windows based on their titles."""
    from fnmatch import fnmatch, fnmatchcase

    existing_types = OrderedSet(win.model_type() for win in ui.iter_windows())

    @configure_gui(
        model_type={"choices": list(existing_types), "value": None},
    )
    def run_collect_windows(
        pattern: str = "*",
        case_sensitive: bool = True,
        model_type: str | None = None,
    ) -> None:
        windows_to_move: list[SubWindow] = []
        _match = fnmatchcase if case_sensitive else fnmatch
        target_index = ui.tabs.current_index
        for win in ui.iter_windows():
            if not _match(win.title, pattern):
                continue
            if model_type is None or win.model_type() == model_type:
                windows_to_move.append(win)
        if windows_to_move:
            for win in windows_to_move:
                ui.move_window(win, target_index)
        ui.tabs.current_index = target_index
        return None

    return run_collect_windows


@ACTIONS.append_from_fn(
    id="close-all-windows",
    title="Close All Windows In Tab",
    menus=[{"id": MenuId.VIEW, "group": WINDOW_GROUP}],
    enablement=_ctx.num_sub_windows > 0,
)
def close_all_windows_in_tab(ui: MainWindow) -> None:
    """Close all sub-windows in the current tab."""
    if area := ui.tabs.current():
        win_modified = [win for win in area if win._need_ask_save_before_close()]
        if len(win_modified) > 0 and ui._instructions.confirm:
            _modified_msg = "\n".join([f"- {win.title}" for win in win_modified])
            ans = ui.exec_choose_one_dialog(
                "Close the tab?",
                f"Some windows in this tab are not saved yet:\n{_modified_msg}\nClose "
                "the tab without saving?",
                ["Yes, close all", "No"],
            )
            if ans == "No":
                raise Cancelled
        area.clear()
    return None


@ACTIONS.append_from_fn(
    id="window-layout-horizontal",
    title="Horizontal Layout ...",
    enablement=(_ctx.num_sub_windows > 1) & (_ctx.num_tabs > 0),
    menus=[MenuId.VIEW_LAYOUT],
    need_function_callback=True,
)
def window_layout_horizontal(ui: MainWindow) -> Parametric:
    """Create a horizontal layout"""
    windows = [win for win in ui.tabs.current()]

    @configure_gui(
        show_parameter_labels=False, wins={"layout": "horizontal", "value": windows[:4]}
    )
    def run_layout_horizontal(wins: list[SubWindow]) -> None:
        layout = ui.tabs.current().add_hbox_layout()
        layout.extend(wins)

    return run_layout_horizontal


@ACTIONS.append_from_fn(
    id="window-layout-vertical",
    title="Vertical Layout ...",
    enablement=(_ctx.num_sub_windows > 1) & (_ctx.num_tabs > 0),
    menus=[MenuId.VIEW_LAYOUT],
    need_function_callback=True,
)
def window_layout_vertical(ui: MainWindow) -> Parametric:
    """Create a vertical layout"""
    windows = [win for win in ui.tabs.current()]

    @configure_gui(
        show_parameter_labels=False, wins={"layout": "vertical", "value": windows[:4]}
    )
    def run_layout_vertical(wins: list[SubWindow]) -> None:
        layout = ui.tabs.current().add_vbox_layout()
        layout.extend(wins)

    return run_layout_vertical


SUBMENUS.append_from(
    id=MenuId.VIEW,
    submenu=MenuId.VIEW_LAYOUT,
    title="Layout",
    group=LAYOUT_GROUP,
)



from __future__ import annotations

from concurrent.futures import Future
from typing import TypeVar, TYPE_CHECKING
from logging import getLogger
import weakref
from magicgui.widgets import FunctionGui
from himena.consts import ParametricWidgetProtocolNames as PWPN
from himena.types import (
    Parametric,
    WidgetType,
    WidgetConstructor,
    WidgetDataModel,
    ClipboardDataModel,
    ParametricWidgetProtocol,
)
from himena.widgets._widget_list import TabArea
from himena.widgets._wrapper import SubWindow
from himena._utils import get_gui_config, ModelTrack
from himena.workflow import CommandExecution

if TYPE_CHECKING:
    from himena._app_model import HimenaApplication
    from himena.widgets._main_window import MainWindow

_W = TypeVar("_W")  # backend widget type

_APP_INSTANCES: dict[str, list[MainWindow]] = {}
_LOGGER = getLogger(__name__)


def current_instance(name: str | None = None) -> MainWindow[_W]:
    """Get current instance of the main window (raise if not exists)."""
    if name is None:
        name = next(iter(_APP_INSTANCES))
    return _APP_INSTANCES[name][-1]


def set_current_instance(name: str, instance: MainWindow[_W]) -> None:
    """Set the instance as the current one."""
    if name not in _APP_INSTANCES:
        _APP_INSTANCES[name] = []
    elif instance in _APP_INSTANCES[name]:
        _APP_INSTANCES[name].remove(instance)
    _APP_INSTANCES[name].append(instance)
    return None


def cleanup():
    """Close all instances and clear the list."""
    for instances in _APP_INSTANCES.copy().values():
        for instance in instances:
            instance.close()
    _APP_INSTANCES.clear()


def remove_instance(name: str, instance: MainWindow[_W]) -> None:
    """Remove the instance from the list."""
    if name in _APP_INSTANCES:
        instances = _APP_INSTANCES[name]
        if instance in instances:
            instances.remove(instance)
            instance.model_app.destroy(instance.model_app.name)
        if not instances:
            _APP_INSTANCES.pop(name, None)
        instance.tabs.clear()
        instance.dock_widgets.clear()
    return None


def _app_destroyed(app_name) -> None:
    """Remove the application from the list."""
    _APP_INSTANCES.pop(app_name, None)


_APP_INITIALIZED = weakref.WeakSet["HimenaApplication"]()


def init_application(app: HimenaApplication) -> HimenaApplication:
    """Register actions/menus and setup injection store."""
    from himena._app_model.actions import ACTIONS, SUBMENUS
    from himena.widgets._main_window import MainWindow

    if app in _APP_INITIALIZED:
        return app

    app.register_actions(ACTIONS)
    app.menus.append_menu_items(SUBMENUS)
    app.destroyed.connect(_app_destroyed)
    _subs = ", ".join(menu.title for _, menu in SUBMENUS)
    _LOGGER.info(f"Initialized submenus: {_subs}")

    app.injection_store.namespace = {
        "MainWindow": MainWindow,
        "TabArea": TabArea,
        "SubWindow": SubWindow,
        "WidgetDataModel": WidgetDataModel,
    }

    ### providers ###
    @app.injection_store.mark_provider
    def _current_instance() -> MainWindow:
        _LOGGER.debug("providing for %r", MainWindow.__name__)
        return current_instance(app.name)

    @app.injection_store.mark_provider
    def _current_tab_area() -> TabArea:
        _LOGGER.debug("providing for %r", TabArea.__name__)
        return current_instance(app.name).tabs.current()

    @app.injection_store.mark_provider
    def _current_window() -> SubWindow:
        _LOGGER.debug("providing for %r", SubWindow.__name__)
        ins = current_instance(app.name)
        if area := ins.tabs.current():
            return area.current()
        return None

    @app.injection_store.mark_provider
    def _provide_data_model() -> WidgetDataModel:
        _LOGGER.debug("providing for %r", WidgetDataModel.__name__)
        return current_instance(app.name)._provide_file_output()[0]

    @app.injection_store.mark_provider
    def _get_clipboard_data() -> ClipboardDataModel:
        _LOGGER.debug("providing for %r", ClipboardDataModel.__name__)
        return current_instance(app.name).clipboard

    ### processors ###
    @app.injection_store.mark_processor
    def _process_data_model(model: WidgetDataModel) -> None:
        if not isinstance(model, WidgetDataModel):
            raise TypeError(f"Expected WidgetDataModel, got {type(model)}")
        ins = current_instance(app.name)
        if ins._instructions.process_model_output:
            _LOGGER.debug("processing %r", model)
            ins.add_data_model(model)
        return None

    @app.injection_store.mark_processor
    def _process_data_models(models: list[WidgetDataModel]) -> None:
        _LOGGER.debug("processing %r", models)
        for each in models:
            _process_data_model(each)

    @app.injection_store.mark_processor
    def _process_clipboard_data(clip_data: ClipboardDataModel) -> None:
        if clip_data is None:
            return None
        _LOGGER.debug("processing %r", clip_data)
        # set data to clipboard
        ins = current_instance(app.name)
        ins.clipboard = clip_data
        return None

    @app.injection_store.mark_processor
    def _process_parametric(fn: Parametric) -> None:
        if fn is None:
            return None
        _LOGGER.debug("processing %r", fn)
        ins = current_instance(app.name)
        gui_config = get_gui_config(fn)
        win = ins.add_function(fn, **gui_config)
        if win._is_run_immediately() and ins._instructions.gui_execution:
            try:
                win._widget_callback()
            finally:
                if win.is_alive:
                    win._close_me(ins)
        else:
            if (tracker := ModelTrack.get(fn)) and tracker.command_id:
                win._widget_workflow = CommandExecution(
                    command_id=tracker.command_id,
                    contexts=tracker.contexts,
                ).construct_workflow()
        return None

    @app.injection_store.mark_processor
    def _process_widget_type(widget: WidgetType) -> None:
        _LOGGER.debug("processing %r", widget)
        ins = current_instance(app.name)
        ins.add_widget(widget)
        return None

    @app.injection_store.mark_processor
    def _process_widget_constructor(con: WidgetConstructor) -> None:
        _LOGGER.debug("processing %r", con)
        ins = current_instance(app.name)
        ins.add_widget(con())
        return None

    @app.injection_store.mark_processor
    def _process_parametric_widget_protocol(widget: ParametricWidgetProtocol) -> None:
        if widget is None:
            return None
        _LOGGER.debug("processing %r", widget)
        ins = current_instance(app.name)
        if isinstance(widget, FunctionGui):
            widget.get_params = widget.asdict
            ins.add_parametric_widget(
                widget,
                callback=widget.__call__,
                title=widget.label,
                preview=False,
                auto_close=False,
                auto_size=False,
            )
        else:
            ins.add_parametric_widget(
                widget,
                title=getattr(widget, PWPN.GET_TITLE, lambda: None)(),
                preview=(
                    hasattr(widget, PWPN.IS_PREVIEW_ENABLED)
                    and hasattr(widget, PWPN.CONNECT_CHANGED_SIGNAL)
                ),
                auto_close=getattr(widget, PWPN.GET_AUTO_CLOSE, lambda: True)(),
                auto_size=getattr(widget, PWPN.GET_AUTO_SIZE, lambda: True)(),
            )
        return None

    @app.injection_store.mark_processor
    def _process_future(future: Future) -> None:
        ui = current_instance(app.name)
        if ui._instructions.unwrap_future:
            app._future_done_callback(future)
        else:
            cb = ui._backend_main_window._process_future_done_callback(
                app._future_done_callback, lambda *_: None
            )
            future.add_done_callback(cb)
            app._futures.add(future)
        return None

    _APP_INITIALIZED.add(app)
    return app



from typing import Sequence, TYPE_CHECKING

import numpy as np
from pydantic import Field
from himena.standards.plotting.layout import BaseLayoutModel
from himena.standards.plotting.components import (
    Axis,
    AxesBase,
    parse_edge,
    parse_face_edge,
)
from himena.standards.plotting import models3d as _m3d

if TYPE_CHECKING:
    from numpy.typing import NDArray


class Axes3D(AxesBase):
    """Layout model for 3D axes."""

    x: Axis = Field(default_factory=Axis, description="X-axis settings.")
    y: Axis = Field(default_factory=Axis, description="Y-axis settings.")
    z: Axis = Field(default_factory=Axis, description="Z-axis settings.")

    def scatter(
        self,
        x: Sequence[float],
        y: Sequence[float],
        z: Sequence[float],
        *,
        symbol: str = "o",
        size: float | None = None,
        **kwargs,
    ) -> _m3d.Scatter3D:
        """Create a 3D scatter plot."""
        model = _m3d.Scatter3D(
            x=x, y=y, z=z, symbol=symbol, size=size, **parse_face_edge(kwargs)
        )
        self.models.append(model)
        return model

    def plot(
        self,
        x: Sequence[float],
        y: Sequence[float],
        z: Sequence[float],
        **kwargs,
    ) -> _m3d.Line3D:
        """Create a 3D line plot."""
        model = _m3d.Line3D(x=x, y=y, z=z, **parse_edge(kwargs))
        self.models.append(model)
        return model

    def surface(
        self,
        x: "NDArray[np.number]",
        y: "NDArray[np.number]",
        z: "NDArray[np.number]",
        **kwargs,
    ) -> _m3d.Surface3D:
        """Create a 3D surface plot."""
        model = _m3d.Surface3D(x=x, y=y, z=z, **parse_face_edge(kwargs))
        self.models.append(model)
        return model

    def mesh(
        self,
        vertices: "NDArray[np.number]",
        face_indices: "NDArray[np.number]",
        **kwargs,
    ) -> _m3d.Mesh3D:
        """Create a 3D mesh plot."""
        model = _m3d.Mesh3D(
            vertices=vertices, face_indices=face_indices, **parse_face_edge(kwargs)
        )
        self.models.append(model)
        return model


class SingleAxes3D(BaseLayoutModel):
    axes: Axes3D = Field(default_factory=Axes3D, description="Child 3D axes.")



from matplotlib import pyplot as plt
from qtpy.QtCore import Qt, QPoint
from pytestqt.qtbot import QtBot
from himena import plotting as hplt, create_model, StandardType
from himena.standards.model_meta import DimAxis
from himena.testing import WidgetTester
from himena_builtins.qt.plot._canvas import QMatplotlibCanvas, QModelMatplotlibCanvas, QModelMatplotlibCanvasStack

def test_matplotlib_canvas(qtbot: QtBot):
    plt.switch_backend("Agg")
    fig = plt.figure()
    canvas = QMatplotlibCanvas()
    qtbot.addWidget(canvas)
    canvas.update_model(create_model(fig, type=StandardType.PLOT))
    with WidgetTester(canvas) as tester:
        tester.to_model()
    canvas._make_context_menu()
    canvas._copy_canvas()
    qtbot.mouseDClick(canvas, Qt.MouseButton.LeftButton, pos=QPoint(10, 10))
    qtbot.mousePress(canvas, Qt.MouseButton.LeftButton, pos=QPoint(10, 10))
    qtbot.mouseMove(canvas, pos=QPoint(20, 20))
    qtbot.mouseRelease(canvas, Qt.MouseButton.LeftButton, pos=QPoint(20, 20))
    plt.close(fig)

def test_model_matplotlib_canvas_single(qtbot: QtBot):
    fig = hplt.figure()
    canvas = QModelMatplotlibCanvas()
    canvas.update_model(create_model(fig, type=StandardType.PLOT))
    qtbot.addWidget(canvas)
    with WidgetTester(canvas) as tester:
        tester.update_model(value=fig, type=StandardType.PLOT)
        tester.cycle_model()
    canvas._make_context_menu()
    canvas._copy_canvas()

def test_model_matplotlib_canvas_row(qtbot: QtBot):
    fig = hplt.row(2)
    fig[0].plot([1, 2, 3])
    fig[1].scatter([4, 5, 6])
    canvas = QModelMatplotlibCanvas()
    canvas.update_model(create_model(fig, type=StandardType.PLOT))
    qtbot.addWidget(canvas)
    with WidgetTester(canvas) as tester:
        tester.cycle_model()

def test_model_matplotlib_canvas_col(qtbot: QtBot):
    fig = hplt.column(2)
    fig[0].bar([1, 2, 3])
    fig[1].errorbar([4, 5, 6])
    canvas = QModelMatplotlibCanvas()
    canvas.update_model(create_model(fig, type=StandardType.PLOT))
    qtbot.addWidget(canvas)
    with WidgetTester(canvas) as tester:
        tester.cycle_model()

def test_model_matplotlib_canvas_stack(qtbot: QtBot):
    fig = hplt.figure_stack(2, 2)
    assert fig.shape == (2, 2)
    fig.axes[0, 0].plot([1, 2, 3])
    fig.axes[0, 1].plot([4, 5, 6])
    fig[1, 0].plot([7, 8, 9])
    fig[1, 1].plot([10, 11, 12])
    fig.x.label = "x"
    fig.y.label = "y"
    fig.title = "title"
    fig.axis_color = "pink"
    canvas = QModelMatplotlibCanvasStack()
    canvas.update_model(create_model(fig, type=StandardType.PLOT_STACK))
    qtbot.addWidget(canvas)
    with WidgetTester(canvas) as tester:
        tester.cycle_model()

    fig = hplt.figure_stack(4, multi_dims="time")
    assert fig.shape == (4,)
    fig = hplt.figure_stack(
        2, 2,
        multi_dims=[
            "time",
            DimAxis(name="slice", scale=0.5, unit="mm", labels=["S0", "S1"]),
        ]
    )
    assert fig.shape == (2, 2)



from __future__ import annotations

from functools import lru_cache
from pathlib import Path
import sys
import subprocess
from abc import ABC, abstractmethod


class BaseInstallTool(ABC):
    """Abstract base class for installation tools like pip and uv."""

    @abstractmethod
    def args_install(self, packages: list[str]) -> list[str]:
        raise NotImplementedError

    def install(self, packages: list[str]) -> subprocess.CompletedProcess:
        return subprocess.run(self.args_install(packages), check=True)


class PipInstallTool(BaseInstallTool):
    def args_install(self, packages: list[str]) -> list[str]:
        return [sys.executable, "-m", "pip", "install", "-U", *packages]


class UvInstallTool(BaseInstallTool):
    def args_install(self, packages: list[str]) -> list[str]:
        return ["uv", "pip", "install", "-U", *packages]


@lru_cache(maxsize=1)
def get_install_tool() -> BaseInstallTool:
    """Get the appropriate installation tool based on the environment."""
    env_dir = Path(sys.executable).parent.parent
    if (venv_cfg_path := env_dir.joinpath("pyvenv.cfg")).exists():
        venv_cfg_dict = _parse_venv_cfg(venv_cfg_path)
        if "uv" in venv_cfg_dict:
            return UvInstallTool()
    return PipInstallTool()


def _parse_venv_cfg(path: Path) -> dict[str, str]:
    cfg = {}
    for line in path.read_text().splitlines():
        if "=" in line:
            key, value = line.split("=", 1)
            cfg[key.strip()] = value.strip()
    return cfg



from __future__ import annotations

from enum import Enum, auto
from contextlib import suppress
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Iterator, cast
from logging import getLogger

from app_model.types import MenuItem
from app_model.backends.qt import QCommandRuleAction
from qtpy import QtWidgets as QtW
from qtpy import QtCore, QtGui
from qtpy.QtCore import Qt
from superqt import QIconifyIcon
from superqt.utils import qthrottled

from himena.qt._qtooltip_widget import QToolTipWidget
from himena.workflow import LocalReaderMethod
from himena._utils import get_display_name
from himena.utils.misc import lru_cache
from himena.consts import MenuId
from himena.types import DragDataModel, WindowState, WindowRect, Size
from himena.plugins import _checker, get_config
from himena.utils.window_rect import ResizeState
from himena.qt._utils import get_main_window, build_qmodel_menu
from himena.qt._qrename import QRenameLineEdit
from himena import _drag

if TYPE_CHECKING:
    from PyQt6 import QtWidgets as QtW
    from himena.qt.main_window import MainWindowQt
    from himena.qt._qmain_window import QMainWindow
    from himena.widgets import SubWindow

_LOGGER = getLogger(__name__)

PROP_IS_CURRENT = "isCurrent"


class QSubWindowArea(QtW.QMdiArea):
    area_focused = QtCore.Signal()

    def __init__(self):
        super().__init__()
        self.viewport().setFocusPolicy(Qt.FocusPolicy.ClickFocus)
        self._last_press_pos: QtCore.QPoint | None = None
        self._last_drag_pos: QtCore.QPoint | None = None
        self.setActivationOrder(QtW.QMdiArea.WindowOrder.ActivationHistoryOrder)
        self._last_active_window_id: QSubWindow | None = None
        self._tooltip_widget = QToolTipWidget(self)

    def addSubWindow(self, sub_window: QSubWindow):
        super().addSubWindow(sub_window)
        sub_window.show()

    def _qmain_window(self) -> QMainWindow | None:
        try:
            return get_main_window(self)._backend_main_window
        except Exception:
            return None

    def iter_widgets(self) -> Iterator[QtW.QWidget]:
        """Iterate over all widgets in the sub-window area."""
        for sub_window in self.subWindowList():
            yield sub_window.widget()

    def indexOf(self, sub_window: QSubWindow) -> int:
        return self.subWindowList().index(sub_window)

    def relabel_widgets(self):
        """Update the 0, 1, 2... labels in the sub-windows."""
        for i, sub_window in enumerate(self.subWindowList()):
            sub_window._title_bar._index_label.setText(str(i))

    def add_widget(
        self,
        widget: QtW.QWidget,
        title: str | None = None,
    ) -> QSubWindow:
        if title is None:
            title = widget.objectName() or "Window"
        if not isinstance(widget, QtW.QWidget):
            raise TypeError(
                f"`widget` must be a QtW.QWidget instance, got {type(widget)}."
            )
        size = widget.sizeHint().expandedTo(QtCore.QSize(160, 120)) + QtCore.QSize(8, 8)
        sub_window = QSubWindow(widget, title)
        self.addSubWindow(sub_window)
        sub_window.resize(size)
        self.relabel_widgets()
        return sub_window

    def mousePressEvent(self, event: QtGui.QMouseEvent):
        self._last_drag_pos = self._last_press_pos = event.pos()
        if (
            event.buttons() & Qt.MouseButton.LeftButton
            and event.modifiers() & Qt.KeyboardModifier.ControlModifier
        ):
            self.setCursor(Qt.CursorShape.ClosedHandCursor)

    def mouseMoveEvent(self, event: QtGui.QMouseEvent):
        self._mouse_move_event(event)

    def _mouse_move_event(self, event: QtGui.QMouseEvent):
        if (
            (event.buttons() & Qt.MouseButton.LeftButton)
            and (main := self._qmain_window())
            and (Qt.Key.Key_Space in main._keys_down)
            or (event.buttons() & Qt.MouseButton.MiddleButton)
        ):
            if self._last_drag_pos is None:
                return None
            # move all the windows
            dpos = event.pos() - self._last_drag_pos
            for sub_window in self.subWindowList():
                sub_window.move(sub_window.pos() + dpos)
            self.setCursor(Qt.CursorShape.ClosedHandCursor)

        self._last_drag_pos = event.pos()

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent):
        # reset cursor state
        is_click = self._last_press_pos is not None and (
            (event.pos() - self._last_press_pos).manhattanLength() < 8
        )
        self.setCursor(Qt.CursorShape.ArrowCursor)
        self._last_press_pos = self._last_drag_pos = None

        # check if any window is under the cursor
        for sub_window in self.subWindowList():
            if sub_window.rect().contains(sub_window.mapFromParent(event.pos())):
                if not sub_window.is_current():
                    sub_window.set_is_current(True)
                    self.subWindowActivated.emit(sub_window)
                break
        else:
            if event.button() == Qt.MouseButton.RightButton and is_click:
                # context menu
                app = get_main_window(self).model_app
                menu = build_qmodel_menu(MenuId.FILE_NEW, app, self)
                menu.exec(event.globalPos())

    def eventFilter(self, obj, a0: QtCore.QEvent) -> bool:
        with suppress(RuntimeError):
            tp = a0.type()
            if tp == QtCore.QEvent.Type.FocusIn:
                if obj is not self:
                    if isinstance(obj, (QtW.QStyle, QtW.QAbstractButton, QtW.QMenuBar)):
                        return False
                    else:
                        if self._last_active_window_id is None and (
                            win := self.currentSubWindow()
                        ):
                            self.subWindowActivated.emit(win)
                            self._last_active_window_id = win._my_wrapper()._identifier
                            _LOGGER.debug("QSubWindowArea.eventFilter: Window focused.")
                else:
                    self._set_area_focused()
                    _LOGGER.debug("QSubWindowArea.eventFilter: TabArea focused.")
            elif tp == QtCore.QEvent.Type.KeyPress:
                a0 = cast(QtGui.QKeyEvent, a0)
                self._set_key_down(a0.key())
            elif tp == QtCore.QEvent.Type.KeyRelease:
                a0 = cast(QtGui.QKeyEvent, a0)
                self._set_key_up(a0.key())
            elif tp == QtCore.QEvent.Type.MouseMove:
                a0 = cast(QtGui.QMouseEvent, a0)
                if self._tooltip_widget.isVisible():
                    if self._tooltip_widget._behavior == "follow":
                        self._tooltip_widget.move_tooltip(QtGui.QCursor.pos())
                    if self._tooltip_widget._behavior == "until_move":
                        self._tooltip_widget.hide()
            return super().eventFilter(obj, a0)
        return False

    def _set_key_down(self, key: int):
        if main := self._qmain_window():
            main._keys_down.add(key)

    def _set_key_up(self, key: int):
        if main := self._qmain_window():
            main._keys_down.discard(key)

    def _set_area_focused(self):
        self.area_focused.emit()
        self._last_active_window_id = None

    def hideEvent(self, a0: QtGui.QHideEvent | None) -> None:
        self._last_drag_pos = self._last_press_pos = None
        return super().hideEvent(a0)

    def _pixmap_resized(
        self,
        size: QtCore.QSize,
        outline: QtGui.QColor | None = None,
    ) -> QtGui.QPixmap:
        return pixmap_resized(self, size, outline)

    if TYPE_CHECKING:

        def subWindowList(self) -> list[QSubWindow]: ...
        def activeSubWindow(self) -> QSubWindow: ...
        def currentSubWindow(self) -> QSubWindow | None: ...


def _get_icon(name: str, rotate=None, color=None):
    try:
        return QIconifyIcon(name, rotate=rotate, color=color)
    except OSError:
        return QtGui.QIcon()


class TitleIconId(Enum):
    WINDOW_MENU = auto()
    MODEL_MENU = auto()
    ACTION_HINT = auto()
    MIN = auto()
    MAX = auto()
    CLOSE = auto()
    NORMAL = auto()


@lru_cache(maxsize=20)
def _icon_for_id(icon_id: TitleIconId, color: str) -> QtGui.QIcon:
    if icon_id is TitleIconId.WINDOW_MENU:
        return _get_icon("material-symbols:menu", color=color)
    elif icon_id is TitleIconId.MODEL_MENU:
        return _get_icon("octicon:ai-model-16", color=color)
    elif icon_id is TitleIconId.ACTION_HINT:
        return _get_icon("ant-design:bulb-outlined", color=color)
    elif icon_id is TitleIconId.MIN:
        return _get_icon("material-symbols:minimize-rounded", color=color)
    elif icon_id is TitleIconId.MAX:
        return _get_icon("material-symbols:crop-5-4-outline", color=color)
    elif icon_id is TitleIconId.CLOSE:
        return _get_icon("material-symbols:close-rounded", color=color)
    elif icon_id is TitleIconId.NORMAL:
        return _get_icon(
            "material-symbols:filter-none-outline-rounded", rotate=180, color=color
        )
    else:
        raise ValueError(f"Invalid icon id: {icon_id}")


class QCentralWidget(QtW.QWidget):
    def __init__(self, parent):
        super().__init__(parent)
        layout = QtW.QVBoxLayout(self)
        layout.setContentsMargins(1, 0, 1, 1)
        layout.setSpacing(0)


class QSubWindow(QtW.QMdiSubWindow):
    state_change_requested = QtCore.Signal(WindowState)
    rename_requested = QtCore.Signal(str)
    close_requested = QtCore.Signal()

    def __init__(self, widget: QtW.QWidget, title: str, icon_color: str = "black"):
        super().__init__()
        self.setWindowFlags(Qt.WindowType.FramelessWindowHint)
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        self.setAttribute(Qt.WidgetAttribute.WA_Hover, True)

        self._window_state = WindowState.NORMAL
        self._resize_state = ResizeState.NONE
        self._widget = widget
        self._current_icon_color = icon_color

        self._central_widget = QCentralWidget(self)
        self.setWidget(self._central_widget)

        self._title_bar = QSubWindowTitleBar(self, title)

        self._central_widget.layout().addWidget(self._title_bar)
        spacer = QtW.QWidget()
        layout = QtW.QVBoxLayout(spacer)
        layout.setContentsMargins(4, 4, 4, 4)
        layout.addWidget(widget)
        self._central_widget.layout().addWidget(spacer)
        self._last_geometry = self.geometry()
        self._anim_geometry = QtCore.QPropertyAnimation(self, b"geometry")
        self.setAcceptDrops(True)

        min_qsize = widget.minimumSizeHint().expandedTo(self._title_bar.minimumSize())
        max_qsize = widget.maximumSize()
        self._min_size = Size(min_qsize.width(), min_qsize.height())
        self._max_size = Size(max_qsize.width(), max_qsize.height())

    def _qt_mdiarea(self) -> QSubWindowArea:
        parent = self
        while parent is not None:
            parent = parent.parentWidget()
            if isinstance(parent, QSubWindowArea):
                return parent
        raise ValueError("Could not find the Qt main window.")

    def windowTitle(self) -> str:
        return self._title_bar._title_label.text()

    def setWindowTitle(self, title: str):
        self._title_bar._title_label.setText(title)

    def _my_wrapper(self) -> SubWindow:
        return self._widget._himena_widget

    @property
    def state(self) -> WindowState:
        return self._window_state

    def _set_icon_color(self, color: str):
        self._title_bar._set_icon_color(color)
        self._current_icon_color = color

    def _update_window_state(self, state: WindowState, animate: bool = True):
        state = WindowState(state)
        g_last = Size(self._last_geometry.width(), self._last_geometry.height())
        if self._window_state == state:
            return None
        if self._qt_mdiarea().viewMode() != QtW.QMdiArea.ViewMode.SubWindowView:
            self._window_state = state
            return None
        if animate:
            _setter = self._set_geometry_animated
        else:
            _setter = self.setGeometry
        g_parent = self.parentWidget().geometry()
        g_size = Size(g_parent.width(), g_parent.height())
        if state == WindowState.MIN:
            if self._window_state is WindowState.NORMAL:
                self._store_current_geometry()
            self.resize(124, self._title_bar.height() + 8)
            n_minimized = sum(
                1
                for sub_window in self._qt_mdiarea().subWindowList()
                if sub_window._window_state is WindowState.MIN
            )
            self._set_minimized(g_parent, n_minimized)
        elif state == WindowState.MAX:
            if self._window_state is WindowState.MIN:
                size_old = g_last
            elif self._window_state is WindowState.NORMAL:
                size_old = Size(self.size().width(), self.size().height())
                self._store_current_geometry()
            else:
                size_old = g_size
            _setter(g_parent)
            self._title_bar._toggle_size_btn.setIcon(
                _icon_for_id(TitleIconId.NORMAL, self._current_icon_color)
            )
            self._widget.setVisible(True)
            _checker.call_widget_resized_callback(
                self._my_wrapper().widget, size_old, g_size
            )
        elif state == WindowState.NORMAL:
            if self._window_state is WindowState.MIN:
                size_old = g_last
            else:
                size_old = g_size
            _setter(self._last_geometry)
            self._title_bar._toggle_size_btn.setIcon(
                _icon_for_id(TitleIconId.MAX, self._current_icon_color)
            )
            self._widget.setVisible(True)
            self._title_bar._fix_position()
            _checker.call_widget_resized_callback(
                self._my_wrapper().widget, size_old, g_last
            )
        elif state == WindowState.FULL:
            if self._window_state is WindowState.MIN:
                size_old = g_last
            elif self._window_state is WindowState.NORMAL:
                size_old = Size(self.size().width(), self.size().height())
                self._store_current_geometry()
            else:
                size_old = g_size
            _setter(g_parent)
            _checker.call_widget_resized_callback(
                self._my_wrapper().widget, size_old, g_last
            )
        else:  # pragma: no cover
            raise RuntimeError(f"Invalid window state value: {state}")

        self._title_bar._window_menu_btn.setVisible(state is not WindowState.MIN)
        self._title_bar._model_menu_btn.setVisible(state is not WindowState.MIN)
        self._title_bar.setVisible(state is not WindowState.FULL)
        self._title_bar._minimize_btn.setVisible(state is not WindowState.MIN)
        self._widget.setVisible(state is not WindowState.MIN)
        self._window_state = state
        return None

    def _store_current_geometry(self):
        self._last_geometry = self.geometry()
        _LOGGER.debug("Storing current geometry %r", self._last_geometry)

    def _set_minimized(self, geometry: QtCore.QRect, number: int = 0):
        self.move(2, geometry.height() - (self._title_bar.height() + 8) * (number + 1))
        self._title_bar._toggle_size_btn.setIcon(
            _icon_for_id(TitleIconId.NORMAL, self._current_icon_color)
        )
        self._widget.setVisible(False)

    def is_current(self) -> bool:
        return self._title_bar.property(PROP_IS_CURRENT)

    def set_is_current(self, is_current: bool):
        """Set the isCurrent state of the sub-window and update styles."""
        self._title_bar.setProperty(PROP_IS_CURRENT, is_current)
        self._title_bar.style().unpolish(self._title_bar)
        self._title_bar.style().polish(self._title_bar)

    def event(self, a0: QtCore.QEvent) -> bool:
        if a0.type() == QtCore.QEvent.Type.HoverMove:
            self._mouse_hover_event(a0.pos())
        elif a0.type() == QtCore.QEvent.Type.MouseButtonPress:
            assert isinstance(a0, QtGui.QMouseEvent)
            self._resize_state = self._check_resize_state(a0.pos())
        elif a0.type() == QtCore.QEvent.Type.MouseButtonRelease:
            self._resize_state = ResizeState.NONE
        return super().event(a0)

    def _check_resize_state(
        self, mouse_pos: QtCore.QPoint, thickness: int = 6, thickness_top: int = 4
    ) -> ResizeState:
        is_left = mouse_pos.x() < thickness
        is_right = mouse_pos.x() > self.width() - thickness
        is_top = mouse_pos.y() < thickness_top
        is_bottom = mouse_pos.y() > self.height() - thickness
        return ResizeState.from_bools(is_left, is_right, is_top, is_bottom)

    @qthrottled(timeout=10)
    def _mouse_hover_event(self, event_pos: QtCore.QPoint):
        # if the cursor is at the edges, set the cursor to resize
        if self._window_state is not WindowState.NORMAL:
            return None
        resize_state = self._check_resize_state(event_pos)
        current_button = QtW.QApplication.mouseButtons()
        if current_button == Qt.MouseButton.NoButton:
            self.setCursor(CURSOR_SHAPE_MAP[resize_state])
        elif current_button & Qt.MouseButton.LeftButton:
            # NOTE: Method "minimusSizeHint" represents the minimum size of the widget
            # as a window
            rect = self.geometry()
            if new_rect := self._resize_state.resize_widget(
                WindowRect(rect.left(), rect.top(), rect.width(), rect.height()),
                mouse_pos=(event_pos.x(), event_pos.y()),
                min_size=self._min_size,
                max_size=self._max_size,
            ):
                subwin = self._my_wrapper()
                size_old = Size(rect.width(), rect.height())

                # update window rect
                subwin.rect = new_rect

                # update window anchor
                main_qsize = self._qt_mdiarea().size()
                subwin.anchor = subwin.anchor.update_for_window_rect(
                    (main_qsize.width(), main_qsize.height()),
                    new_rect,
                )
                _checker.call_widget_resized_callback(
                    self._my_wrapper().widget,
                    size_old,
                    new_rect.size(),
                )
        return None

    def _set_geometry_animated(self, rect: QtCore.QRect):
        if self._anim_geometry.state() == QtCore.QAbstractAnimation.State.Running:
            self._anim_geometry.stop()
        self._anim_geometry.setTargetObject(self)
        self._anim_geometry.setPropertyName(b"geometry")
        self._anim_geometry.setStartValue(QtCore.QRect(self.geometry()))
        self._anim_geometry.setEndValue(rect)
        self._anim_geometry.setDuration(60)
        self._anim_geometry.start()

    def _pixmap_resized(
        self,
        size: QtCore.QSize,
        outline: QtGui.QColor | None = None,
    ) -> QtGui.QPixmap:
        return pixmap_resized(self, size, outline)

    def _find_me(self) -> tuple[int, int]:
        return self._find_me_and_main()[0]

    def _find_me_and_main(self) -> tuple[tuple[int, int], MainWindowQt]:
        main = get_main_window(self)
        for i_tab, tab in main.tabs.enumerate():
            for i_win, win in tab.enumerate():
                _, qwidget = win._split_interface_and_frontend()
                if qwidget is self._widget:
                    return (i_tab, i_win), main
        raise RuntimeError("Could not find the sub-window in the main window.")

    def dragEnterEvent(self, a0: QtGui.QDragEnterEvent | None) -> None:
        if a0 is None:
            return None
        if model := _drag.get_dragging_model():
            if self._my_wrapper()._is_drop_accepted(model):
                a0.accept()
                return None
        a0.ignore()

    def dropEvent(self, a0: QtGui.QDropEvent | None) -> None:
        if a0 is None:
            return None
        if model := _drag.drop():
            # user defined drop event
            self_wrapper = self._my_wrapper()
            if isinstance(src := a0.source(), QSubWindow):
                source = src._my_wrapper()
            else:
                source = None
            if self_wrapper.widget is src:  # same widget
                a0.ignore()
                _LOGGER.info("dropping to the same widget, ignore it")
                return None

            if self_wrapper._process_drop_event(model, source):
                a0.accept()
                return None
        a0.ignore()
        a0.setDropAction(Qt.DropAction.IgnoreAction)

    def focusOutEvent(self, e):
        self._qt_mdiarea()._tooltip_widget.hide()
        return super().focusOutEvent(e)


class QTitleBarToolButton(QtW.QToolButton):
    """Tool button for the title bar of the sub-window."""

    def __init__(
        self,
        icon_id: TitleIconId,
        color: str,
        tooltip: str,
        callback: Callable[[], None],
    ):
        super().__init__()
        self._icon_id = icon_id
        self._current_icon_color = color
        self.setToolTip(tooltip)
        self.clicked.connect(callback)
        self._update_icon_color(color)

    def _set_size(self, size: int):
        self.setFixedSize(size, size)
        self.setIconSize(QtCore.QSize(size - 1, size - 1))

    def _update_icon_color(self, color: str):
        self._current_icon_color = color
        self.setIcon(_icon_for_id(self._icon_id, color))


class QDummyToolButton(QtW.QWidget):
    """Invisible widget just for API compatibility."""

    def __init__(self):
        super().__init__()
        self.setVisible(False)

    def _set_size(self, size: int):
        self.setFixedSize(size, size)

    def _update_icon_color(self, color: str):
        """Do nothing."""


class QSubWindowTitleBar(QtW.QFrame):
    """The subwindow title bar widget."""

    def __init__(self, subwindow: QSubWindow, title: str):
        super().__init__()
        self._drag_position: QtCore.QPoint | None = None
        self._is_window_drag_mode: bool = False
        self._is_double_clicking: bool = False
        self._resize_position: QtCore.QPoint | None = None
        self._subwindow = subwindow

        self.setFrameShape(QtW.QFrame.Shape.StyledPanel)
        self.setFrameShadow(QtW.QFrame.Shadow.Raised)
        self.setMinimumWidth(100)
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)

        self._window_menu_btn = QTitleBarToolButton(
            icon_id=TitleIconId.WINDOW_MENU,
            color="black",
            tooltip="Menu for this window",
            callback=self._show_window_menu,
        )
        if self._get_model_type() is not None:
            self._model_menu_btn = QTitleBarToolButton(
                icon_id=TitleIconId.MODEL_MENU,
                color="black",
                tooltip="Menu specific to the model",
                callback=self._show_model_menu,
            )
            # action hint button
            self._action_hint_btn = QTitleBarToolButton(
                icon_id=TitleIconId.ACTION_HINT,
                color="black",
                tooltip="Click to see available action hints",
                callback=self._show_action_hints,
            )
        else:
            self._model_menu_btn = QDummyToolButton()
            self._action_hint_btn = QDummyToolButton()

        self._index_label = QtW.QLabel()
        self._index_label.setObjectName("indexLabel")
        self._index_label.setAlignment(
            Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignVCenter
        )
        _index_font = self._index_label.font()
        _index_font.setPointSize(8)
        _index_font.setBold(True)
        self._index_label.setFont(_index_font)
        self._index_label.setFixedWidth(20)
        self._index_label.setContentsMargins(0, 0, 0, 0)
        self._index_label.setSizePolicy(
            QtW.QSizePolicy.Policy.Fixed, QtW.QSizePolicy.Policy.Fixed
        )

        self._title_label = QtW.QLabel(title)
        self._title_label.setIndent(3)
        self._title_label.setContentsMargins(0, 0, 0, 0)
        self._title_label.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Fixed
        )

        self._line_edit = QRenameLineEdit(self._title_label)

        @self._line_edit.rename_requested.connect
        def _(new_name: str):
            self._subwindow.rename_requested.emit(new_name)

        self._minimize_btn = QTitleBarToolButton(
            icon_id=TitleIconId.MIN,
            color="black",
            tooltip="Minimize this window",
            callback=self._minimize,
        )
        self._toggle_size_btn = QTitleBarToolButton(
            icon_id=TitleIconId.MAX,
            color="black",
            tooltip="Toggle the size of this window",
            callback=self._toggle_size,
        )
        self._close_btn = QTitleBarToolButton(
            icon_id=TitleIconId.CLOSE,
            color="black",
            tooltip="Close this window",
            callback=self._close,
        )

        layout = QtW.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(1)
        layout.addWidget(self._window_menu_btn)
        layout.addWidget(self._model_menu_btn)
        layout.addWidget(self._action_hint_btn)
        layout.addWidget(self._index_label)
        layout.addWidget(self._title_label)
        layout.addWidget(self._minimize_btn, alignment=Qt.AlignmentFlag.AlignRight)
        layout.addWidget(self._toggle_size_btn, alignment=Qt.AlignmentFlag.AlignRight)
        layout.addWidget(self._close_btn, alignment=Qt.AlignmentFlag.AlignRight)

        self.setProperty(PROP_IS_CURRENT, False)
        self.setAcceptDrops(True)

        from himena.plugins.install import GlobalConfig

        cfg = get_config(GlobalConfig) or GlobalConfig()
        self._bar_size = cfg.subwindow_bar_height
        self._set_bar_size(self._bar_size)

    def _get_model_type(self) -> str | None:
        interf = self._subwindow._my_wrapper().widget
        if hasattr(interf, "model_type"):
            try:
                return interf.model_type()
            except Exception:
                return None
        elif hasattr(interf, "__himena_model_type__"):
            return interf.__himena_model_type__
        return None

    def _set_visible(self, visible: bool):
        if visible:
            self._set_bar_size(self._bar_size)
        else:
            self._set_bar_size(0)

    def _set_bar_size(self, height: int):
        self.setFixedHeight(height)
        self._index_label.setFixedHeight(height)
        self._title_label.setFixedHeight(height)
        if height > 1:
            self._window_menu_btn._set_size(height - 1)
            self._model_menu_btn._set_size(height - 1)
            self._action_hint_btn._set_size(height - 1)
            self._minimize_btn._set_size(height - 1)
            self._toggle_size_btn._set_size(height - 1)
            self._close_btn._set_size(height - 1)

    def _set_icon_color(self, color):
        for toolbtn in (
            self._window_menu_btn,
            self._model_menu_btn,
            self._action_hint_btn,
            self._minimize_btn,
            self._toggle_size_btn,
            self._close_btn,
        ):
            toolbtn._update_icon_color(color)

    def _start_renaming(self):
        self._line_edit.show()
        self._move_line_edit(self._title_label.rect(), self._title_label.text())

    def _move_line_edit(
        self,
        rect: QtCore.QRect,
        text: str,
    ) -> QtW.QLineEdit:
        geometry = self._line_edit.geometry()
        geometry.setWidth(rect.width())
        geometry.setHeight(rect.height())
        geometry.moveCenter(rect.center())
        self._line_edit.setGeometry(geometry)
        self._line_edit.setText(text)
        self._line_edit.setHidden(False)
        self._line_edit.setFocus()
        self._line_edit.selectAll()

    def _minimize(self):
        self._subwindow.state_change_requested.emit(WindowState.MIN)

    def _toggle_size(self):
        if self._subwindow._window_state is WindowState.NORMAL:
            self._subwindow.state_change_requested.emit(WindowState.MAX)
        else:
            self._subwindow.state_change_requested.emit(WindowState.NORMAL)

    def _maximize(self):
        self._subwindow.state_change_requested.emit(WindowState.MAX)

    def _toggle_full_screen(self):
        if self._subwindow._window_state is WindowState.FULL:
            self._subwindow.state_change_requested.emit(WindowState.NORMAL)
        else:
            self._subwindow.state_change_requested.emit(WindowState.FULL)

    def _close(self):
        return self._subwindow.close_requested.emit()

    def _make_tooltip(self):
        """Make the tooltip for the title bar"""
        qwin = self._subwindow
        attrs: list[str] = [f"<b>Title</b>: {self._title_label.text()}"]
        if _model_type := self._get_model_type():
            attrs.append(f"<b>Type</b>: {_model_type}")
        attrs.append(
            f"<b>Widget</b>: {get_display_name(qwin._widget.__class__, sep=' ')}"
        )
        sub = qwin._my_wrapper()
        attrs.append(f"<b>Save behavior</b>: {sub.save_behavior!r}")
        tooltip = "<br>".join(attrs)
        return tooltip

    def event(self, a0: QtCore.QEvent) -> bool:
        if a0.type() == QtCore.QEvent.Type.ToolTip:
            a0 = cast(QtGui.QHelpEvent, a0)
            QtW.QToolTip.showText(a0.globalPos(), self._make_tooltip(), self)
            return True
        return super().event(a0)

    # drag events for moving the window
    def mousePressEvent(self, event: QtGui.QMouseEvent):
        self._is_window_drag_mode = False
        if self._is_double_clicking:
            return super().mousePressEvent(event)
        _subwin = self._subwindow
        _left_btn = event.buttons() & Qt.MouseButton.LeftButton
        _middle_btn = event.buttons() & Qt.MouseButton.MiddleButton
        _ctrl_mod = event.modifiers() & Qt.KeyboardModifier.ControlModifier
        if (_left_btn and _ctrl_mod) or _middle_btn:
            # start dragging subwindow
            self._is_window_drag_mode = True
            drag = self._make_subwindow_drag()
            self._subwindow.hide()
            try:
                drag.exec()
            finally:
                # NOTE: if subwindow is dropped to another tab, the old one will be
                # removed from the MdiArea. In this case, the subwindow should not
                # be shown again.
                if self._subwindow.parent():
                    self._subwindow.show()
        else:
            if _subwin._window_state == WindowState.MIN:
                # cannot move minimized window
                return
        self._store_drag_position(event.globalPos())
        return super().mousePressEvent(event)

    def mouseMoveEvent(self, event: QtGui.QMouseEvent):
        _subwin = self._subwindow
        if event.buttons() == Qt.MouseButton.NoButton:
            self._drag_position = None
        elif (
            event.buttons() == Qt.MouseButton.LeftButton
            and self._drag_position is not None
            and _subwin._resize_state is ResizeState.NONE
            and not self._is_window_drag_mode
        ):
            g = _subwin.geometry()
            main_qsize = _subwin._qt_mdiarea().size()
            wrapper = _subwin._my_wrapper()
            if _subwin._window_state == WindowState.MAX:
                # change to normal without moving
                self._toggle_size_btn.setIcon(
                    _icon_for_id(TitleIconId.MAX, self._subwindow._current_icon_color)
                )
                _subwin._widget.setVisible(True)
                _subwin._window_state = WindowState.NORMAL
                # restore the last size
                _last_geo = _subwin._last_geometry
                _next_rect = WindowRect(
                    event.pos().x() - _last_geo.width() // 2,
                    event.pos().y() - self.height() // 2,
                    _last_geo.width(),
                    _last_geo.height(),
                )
                self._store_drag_position(event.globalPos())
            else:
                # drag the subwindow
                new_pos = event.globalPos() - self._drag_position
                offset = self.height() - 4
                if new_pos.y() < -offset:
                    new_pos.setY(-offset)
                _next_rect = wrapper.rect.move_top_left(new_pos.x(), new_pos.y())

            # update window anchor
            wrapper.rect = _next_rect
            wrapper.anchor = wrapper.anchor.update_for_window_rect(
                (main_qsize.width(), main_qsize.height()),
                WindowRect.from_tuple(g.left(), g.top(), g.width(), g.height()),
            )
        return super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent):
        self._is_window_drag_mode = False
        self._is_double_clicking = False
        if event.button() == Qt.MouseButton.LeftButton:
            self._drag_position = None
            self._fix_position()
        elif event.button() == Qt.MouseButton.RightButton:
            # context menu
            if (event.pos() - self._drag_position).manhattanLength() < 10:
                context_menu = self._prep_window_menu()
                context_menu.exec(event.globalPos())
        return super().mouseReleaseEvent(event)

    def mouseDoubleClickEvent(self, event: QtGui.QMouseEvent):
        self._is_double_clicking = True
        if event.buttons() & Qt.MouseButton.LeftButton:
            self._toggle_size()
        return super().mouseDoubleClickEvent(event)

    def wheelEvent(self, event: QtGui.QWheelEvent):
        if event.modifiers() == Qt.KeyboardModifier.NoModifier:
            self._wheel_event(event.angleDelta().y())
        return super().wheelEvent(event)

    def _wheel_event(self, dy: int):
        subwin = self._subwindow
        i_tab, i_win = subwin._find_me()
        main = get_main_window(subwin)
        sub = main.tabs[i_tab][i_win]
        size_old = sub.size
        if dy > 0:
            rect_new = sub.rect.resize_relative(1.1, 1.1)
        else:
            rect_new = sub.rect.resize_relative(1 / 1.1, 1 / 1.1)
        inst = main._instructions.updated(animate=False)
        sub._set_rect(rect_new, inst)
        size_new = rect_new.size()
        return _checker.call_widget_resized_callback(
            subwin._my_wrapper().widget, size_old, size_new
        )

    def _make_subwindow_drag(self):
        _subwin = self._subwindow
        drag = QtGui.QDrag(_subwin)
        drag.destroyed.connect(_drag.clear)
        mime_data = QtCore.QMimeData()
        _wrapper = self._subwindow._my_wrapper()
        if isinstance(
            _meth := _wrapper._widget_workflow.last(),
            LocalReaderMethod,
        ):
            if isinstance(_meth.path, Path):
                mime_data.setUrls([QtCore.QUrl(_meth.path.as_uri())])
            elif isinstance(_meth.path, list):
                mime_data.setUrls([QtCore.QUrl(fp.as_uri()) for fp in _meth.path])
        drag.setMimeData(mime_data)
        if _wrapper.supports_to_model:
            model = DragDataModel(
                getter=_wrapper.to_model,
                type=_wrapper.model_type(),
            )
            _drag.drag(model)
        drag.setPixmap(_subwin._pixmap_resized(QtCore.QSize(150, 150)))
        return drag

    def dragEnterEvent(self, a0: QtGui.QDragEnterEvent | None) -> None:
        return self._subwindow.dragEnterEvent(a0)

    def dropEvent(self, a0: QtGui.QDropEvent | None) -> None:
        return self._subwindow.dropEvent(a0)

    def enterEvent(self, event):
        ui = get_main_window(self)
        ui.set_status_tip("`Ctrl+LeftButton` or `MiddleButton` to drag the window")
        return super().enterEvent(event)

    def leaveEvent(self, a0):
        try:
            main = get_main_window(self)
        except ValueError:
            pass
        else:
            main.set_status_tip("")
        return super().leaveEvent(a0)

    def _store_drag_position(self, global_pos: QtCore.QPoint):
        subwin = self._subwindow
        self._drag_position = global_pos - subwin.frameGeometry().topLeft()

    def _fix_position(self):
        self_pos = self.mapToGlobal(self._subwindow.rect().topRight())
        parent_pos = self._subwindow.parentWidget().mapToGlobal(QtCore.QPoint(0, 0))
        if self_pos.y() < parent_pos.y():
            self._subwindow.move(self._subwindow.pos().x(), 0)
        if self_pos.x() < parent_pos.x():
            self._subwindow.move(0, self._subwindow.pos().y())

    def _show_model_menu(self):
        context_menu = self._prep_model_menu()
        return self._exec_menu_at_button(context_menu, self._model_menu_btn)

    def _show_action_hints(self):
        context_menu = self._prep_action_hints_menu()
        if len(context_menu.actions()) == 0:
            action = context_menu.addAction("No action hints available", lambda: None)
            action.setEnabled(False)
        return self._exec_menu_at_button(context_menu, self._action_hint_btn)

    def _prep_window_menu(self) -> QtW.QMenu:
        main = get_main_window(self)
        app = main._model_app

        context_menu = build_qmodel_menu(MenuId.WINDOW, app=app.name, parent=self)
        ctx = main._ctx_keys
        ctx._update(main)
        context_menu.update_from_context(ctx.dict())
        return context_menu

    def _prep_model_menu(self) -> QtW.QMenu:
        model_type = self._get_model_type()
        if model_type is None:
            return None
        ui = get_main_window(self)
        model_subtypes = model_type.split(".")
        supertype_menus: list[QtW.QMenu] = []
        _id = f"/model_menu:{model_type}"

        context_menu = build_qmodel_menu(_id, app=ui.model_app.name, parent=self)
        ctx = ui._ctx_keys
        ctx._update(ui)
        ctx_dict = ctx.dict()
        open_in_actions: list[QCommandRuleAction] = []
        _n_enabled = 0
        for i in range(1, len(model_subtypes) + 1):
            _typ = ".".join(model_subtypes[:i])
            _id_open_in = f"/open-in/{_typ}"
            for menu_group in ui.model_app.menus.iter_menu_groups(_id_open_in):
                for menu_or_submenu in menu_group:
                    if isinstance(menu_or_submenu, MenuItem):
                        action = QCommandRuleAction(
                            menu_or_submenu.command, ui.model_app.name, self
                        )
                        is_enabled = menu_or_submenu.command.enablement.eval(ctx_dict)
                        _n_enabled += int(is_enabled)
                        action.setEnabled(is_enabled)
                        open_in_actions.append(action)
        if open_in_actions and _n_enabled > 0:
            open_in_menu = QtW.QMenu()
            open_in_menu.setParent(context_menu, open_in_menu.windowFlags())
            open_in_menu.setTitle("Open in ...")
            actions = context_menu.actions()
            if len(actions) == 0:
                context_menu.addMenu(open_in_menu)
            else:
                context_menu.insertMenu(actions[0], open_in_menu)
            for action in open_in_actions:
                open_in_menu.addAction(action)

        # also add supertype actions
        for num in range(1, len(model_subtypes)):
            _typ = ".".join(model_subtypes[:num])
            _id = f"/model_menu:{_typ}"
            _menu = build_qmodel_menu(_id, app=ui.model_app.name, parent=self)
            _menu.setTitle(f'"{_typ}" menu')
            supertype_menus.append(_menu)
        if supertype_menus:
            context_menu.addSeparator()
            for menu in supertype_menus:
                if not menu.isEmpty():
                    context_menu.addMenu(menu)
        context_menu.update_from_context(ctx_dict)
        return context_menu

    def _prep_action_hints_menu(self) -> QtW.QMenu:
        ui = get_main_window(self)
        model_type = self._get_model_type()
        if model_type is None:
            return QtW.QMenu()
        subwin = self._subwindow._my_wrapper()
        last_step = subwin._widget_workflow.last()
        if last_step is None:
            return QtW.QMenu()
        menu = QtW.QMenu()
        for sug in ui.action_hint_registry.iter_suggestion(model_type, last_step):
            menu.addAction(sug.get_title(ui), sug.make_executor(ui, last_step))
        menu.setParent(self, menu.windowFlags())
        return menu

    def _show_window_menu(self):
        context_menu = self._prep_window_menu()
        return self._exec_menu_at_button(context_menu, self._window_menu_btn)

    def _exec_menu_at_button(self, menu: QtW.QMenu, btn: QtW.QToolButton):
        pos_local = btn.rect().bottomLeft()
        pos_global = btn.mapToGlobal(pos_local)
        menu.exec(pos_global)


def pixmap_resized(
    widget: QtW.QWidget,
    size: QtCore.QSize,
    outline: QtGui.QColor | None = None,
) -> QtGui.QPixmap:
    pixmap = widget.grab().scaled(
        size,
        Qt.AspectRatioMode.KeepAspectRatio,
        Qt.TransformationMode.SmoothTransformation,
    )
    if outline is not None:
        painter = QtGui.QPainter(pixmap)
        painter.setPen(QtGui.QPen(outline, 2))
        painter.drawRect(pixmap.rect().adjusted(0, 0, -1, -1))
        painter.end()
    return pixmap


CURSOR_SHAPE_MAP = {
    ResizeState.TOP: Qt.CursorShape.SizeVerCursor,
    ResizeState.BOTTOM: Qt.CursorShape.SizeVerCursor,
    ResizeState.LEFT: Qt.CursorShape.SizeHorCursor,
    ResizeState.RIGHT: Qt.CursorShape.SizeHorCursor,
    ResizeState.TOP_LEFT: Qt.CursorShape.SizeFDiagCursor,
    ResizeState.TOP_RIGHT: Qt.CursorShape.SizeBDiagCursor,
    ResizeState.BOTTOM_LEFT: Qt.CursorShape.SizeBDiagCursor,
    ResizeState.BOTTOM_RIGHT: Qt.CursorShape.SizeFDiagCursor,
    ResizeState.NONE: Qt.CursorShape.ArrowCursor,
}


def get_subwindow(widget: QtW.QWidget) -> QSubWindow:
    window = widget.parentWidget().parentWidget().parentWidget()
    if not isinstance(window, QSubWindow):
        raise ValueError(f"Widget {widget!r} is not in a sub-window.")
    return window
